# -*- coding: utf-8 -*-

import os
import oss2


def select_call_back(consumed_bytes, total_bytes = None):
    print('Consumed Bytes:' + str(consumed_bytes) + '\n')
# 首先初始化AccessKeyId、AccessKeySecret、Endpoint等信息。
# 通过环境变量获取，或者把诸如“<你的AccessKeyId>”替换成真实的AccessKeyId等。
#
# 以杭州区域为例，Endpoint可以是：
#   http://oss-cn-hangzhou.aliyuncs.com
#   https://oss-cn-hangzhou.aliyuncs.com
# 分别以HTTP、HTTPS协议访问。
access_key_id = os.getenv('OSS_TEST_ACCESS_KEY_ID', '<你的AccessKeyId>')
access_key_secret = os.getenv('OSS_TEST_ACCESS_KEY_SECRET', '<你的AccessKeySecret>')
bucket_name = os.getenv('OSS_TEST_BUCKET', '<你的Bucket>')
endpoint = os.getenv('OSS_TEST_ENDPOINT', '<你的访问域名>')
# 确认上面的参数都填写正确了
for param in (access_key_id, access_key_secret, bucket_name, endpoint):
    assert '<' not in param, '请设置参数：' + param


# 创建Bucket对象，所有Object相关的接口都可以通过Bucket对象来进行
bucket = oss2.Bucket(oss2.Auth(access_key_id, access_key_secret), endpoint, bucket_name)
#objects = bucket.list_objects()
key = 'python_select.csv'
content = 'Tom Hanks,USA,45\r\n'*1024
filename = 'python_select.csv'

# 上传文件
bucket.put_object(key, content)

csv_meta_params = {'RecordDelimiter': '\r\n'}

select_csv_params = {'CsvHeaderInfo': 'None',
                'RecordDelimiter': '\r\n',
                'LineRange': (500, 1000)}

csv_header = bucket.create_select_object_meta(key, csv_meta_params)
print(csv_header.csv_rows)
print(csv_header.csv_splits)

result = bucket.select_object(key, "select * from ossobject where _3 > 44 limit 100000", select_call_back, select_csv_params)
content_got = b''
for chunk in result:
   content_got += chunk
print(content_got)
result = bucket.select_object_to_file(key, filename, 
        "select * from ossobject where _3 > 44 limit 100000", select_call_back, select_csv_params)

bucket.delete_object(key)


# -*- coding: utf-8 -*-
"""
oss2.headers
~~~~~~~~
这个模块包含http请求里header的key定义
同时包含了发送http请求的header, 类型为dict
"""
OSS_USER_METADATA_PREFIX = "x-oss-meta-"

OSS_CANNED_ACL = "x-oss-acl"

IF_UNMODIFIED_SINCE = "If-Unmodified-Since"
IF_MATCH = "If-Match"

OSS_COPY_OBJECT_SOURCE = "x-oss-copy-source"
OSS_COPY_OBJECT_SOURCE_RANGE = "x-oss-copy-source-range"

OSS_REQUEST_ID = "x-oss-request-id"

OSS_SECURITY_TOKEN = "x-oss-security-token"

OSS_NEXT_APPEND_POSITION = "x-oss-next-append-position"
OSS_HASH_CRC64_ECMA = "x-oss-hash-crc64ecma"
OSS_OBJECT_TYPE = "x-oss-object-type"

OSS_OBJECT_ACL = "x-oss-object-acl"

OSS_SYMLINK_TARGET = "x-oss-symlink-target"

OSS_SERVER_SIDE_ENCRYPTION = "x-oss-server-side-encryption"
OSS_SERVER_SIDE_ENCRYPTION_KEY_ID = "x-oss-server-side-encryption-key-id"

OSS_CLIENT_SIDE_ENCRYPTION_KEY = "x-oss-meta-client-side-encryption-key"
OSS_CLIENT_SIDE_ENCRYPTION_START = "x-oss-meta-client-side-encryption-start"
OSS_CLIENT_SIDE_ENCRYPTION_CEK_ALG = "x-oss-meta-client-side-encryption-cek-alg"
OSS_CLIENT_SIDE_ENCRYPTION_WRAP_ALG = "x-oss-meta-client-side-encryption-wrap-alg"
OSS_CLIENT_SIDE_ENCRYTPION_MATDESC = "x-oss-meta-client-side-encryption-matdesc"
OSS_CLIENT_SIDE_ENCRYPTION_UNENCRYPTED_CONTENT_LENGTH = "x-oss-meta-client-side-encryption-unencrypted-content-length"
OSS_CLIENT_SIDE_ENCRYPTION_UNENCRYPTED_CONTENT_MD5 = "x-oss-meta-client-side-encryption-unencrypted-content-md5"
OSS_CLIENT_SIDE_ENCRYPTION_DATA_SIZE = "x-oss-meta-client-side-encryption-data-size"
OSS_CLIENT_SIDE_ENCRYPTION_PART_SIZE = "x-oss-meta-client-side-encryption-part-size"

DEPRECATED_CLIENT_SIDE_ENCRYPTION_KEY = "x-oss-meta-oss-crypto-key"
DEPRECATED_CLIENT_SIDE_ENCRYPTION_START = "x-oss-meta-oss-crypto-start"
DEPRECATED_CLIENT_SIDE_ENCRYPTION_CEK_ALG = "x-oss-meta-oss-cek-alg"
DEPRECATED_CLIENT_SIDE_ENCRYPTION_WRAP_ALG = "x-oss-meta-oss-wrap-alg"
DEPRECATED_CLIENT_SIDE_ENCRYTPION_MATDESC = "x-oss-meta-oss-crypto-matdesc"
DEPRECATED_CLIENT_SIDE_ENCRYPTION_UNENCRYPTED_CONTENT_LENGTH = "x-oss-meta-oss-crypto-unencrypted-content-length"
DEPRECATED_CLIENT_SIDE_ENCRYPTION_UNENCRYPTED_CONTENT_MD5 = "x-oss-meta-oss-crypto-unencrypted-content-md5"

OSS_OBJECT_TAGGING = "x-oss-tagging"
OSS_OBJECT_TAGGING_COPY_DIRECTIVE = "x-oss-tagging-directive"

OSS_REQUEST_PAYER = 'x-oss-request-payer'

OSS_TRAFFIC_LIMIT = 'x-oss-traffic-limit'

RSA_NONE_PKCS1Padding_WRAP_ALGORITHM = 'RSA/NONE/PKCS1Padding'
RSA_NONE_OAEPWithSHA1AndMGF1Padding = 'RSA/NONE/OAEPWithSHA-1AndMGF1Padding'
KMS_ALI_WRAP_ALGORITHM = 'KMS/ALICLOUD'
OSS_ENCRYPTION_CLIENT = 'OssEncryptionClient'
OSS_TASK_ID = 'x-oss-task-id'

OSS_SERVER_SIDE_ENCRYPTION = "x-oss-server-side-encryption"
OSS_SERVER_SIDE_ENCRYPTION_KEY_ID = "x-oss-server-side-encryption-key-id"
OSS_SERVER_SIDE_DATA_ENCRYPTION = "x-oss-server-side-data-encryption"

OSS_METADATA_DIRECTIVE = 'x-oss-metadata-directive'

class RequestHeader(dict):
    def __init__(self, *arg, **kw):
        super(RequestHeader, self).__init__(*arg, **kw)

    def set_server_side_encryption(self, algorithm=None, cmk_id=None):
        if OSS_SERVER_SIDE_ENCRYPTION in self:
            del self[OSS_SERVER_SIDE_ENCRYPTION]
        if OSS_SERVER_SIDE_ENCRYPTION_KEY_ID in self:
            del self[OSS_SERVER_SIDE_ENCRYPTION_KEY_ID]

        if algorithm == "AES256":
            self[OSS_SERVER_SIDE_ENCRYPTION] = "AES256"
        elif algorithm == "KMS":
            self[OSS_SERVER_SIDE_ENCRYPTION] = "KMS"
            if cmk_id is not None:
                self[OSS_SERVER_SIDE_ENCRYPTION_KEY_ID] = cmk_id



# -*- coding: utf-8 -*-

"""
oss2.defaults
~~~~~~~~~~~~~

全局缺省变量。

"""


def get(value, default_value):
    if value is None:
        return default_value
    else:
        return value


#: 连接超时时间
connect_timeout = 60

#: 缺省重试次数
request_retries = 3

#: 对于某些接口，上传数据长度大于或等于该值时，就采用分片上传。
multipart_threshold = 10 * 1024 * 1024

#: 分片上传缺省线程数
multipart_num_threads = 1

#: 缺省分片大小
part_size = 10 * 1024 * 1024

#: 最大分片的个数
max_part_count = 10000

#: 最小分片大小
min_part_size = 100 * 1024

#: 每个Session连接池大小
connection_pool_size = 10


#: 对于断点下载，如果OSS文件大小大于该值就进行并行下载（multiget）
multiget_threshold = 100 * 1024 * 1024

#: 并行下载（multiget）缺省线程数
multiget_num_threads = 4

#: 并行下载（multiget）的缺省分片大小
multiget_part_size = 10 * 1024 * 1024



# -*- coding: utf-8 -*-

"""
oss2.exceptions
~~~~~~~~~~~~~~

异常类。
"""

import re

import xml.etree.ElementTree as ElementTree
from xml.parsers import expat


from .compat import to_string
from .headers import *


_OSS_ERROR_TO_EXCEPTION = {}  # populated at end of module


OSS_CLIENT_ERROR_STATUS = -1
OSS_REQUEST_ERROR_STATUS = -2
OSS_INCONSISTENT_ERROR_STATUS = -3
OSS_FORMAT_ERROR_STATUS = -4
OSS_SELECT_CLIENT_ERROR_STATUS = -5


class OssError(Exception):
    def __init__(self, status, headers, body, details):
        #: HTTP 状态码
        self.status = status

        #: 请求ID，用于跟踪一个OSS请求。提交工单时，最好能够提供请求ID
        self.request_id = headers.get(OSS_REQUEST_ID, '')

        #: HTTP响应体（部分）
        self.body = body

        #: 详细错误信息，是一个string到string的dict
        self.details = details

        #: OSS错误码
        self.code = self.details.get('Code', '')

        #: OSS错误信息
        self.message = self.details.get('Message', '')

    def __str__(self):
        error = {'status': self.status,
                 OSS_REQUEST_ID : self.request_id,
                 'details': self.details}
        return str(error)

    def _str_with_body(self):
        error = {'status': self.status,
                 OSS_REQUEST_ID : self.request_id,
                 'details': self.body}
        return str(error)


class ClientError(OssError):
    def __init__(self, message):
        OssError.__init__(self, OSS_CLIENT_ERROR_STATUS, {}, 'ClientError: ' + message, {})

    def __str__(self):
        return self._str_with_body()


class RequestError(OssError):
    def __init__(self, e):
        OssError.__init__(self, OSS_REQUEST_ERROR_STATUS, {}, 'RequestError: ' + str(e), {})
        self.exception = e

    def __str__(self):
        return self._str_with_body()


class InconsistentError(OssError):
    def __init__(self, message, request_id=''):
        OssError.__init__(self, OSS_INCONSISTENT_ERROR_STATUS, {OSS_REQUEST_ID : request_id}, 'InconsistentError: ' + message, {})

    def __str__(self):
        return self._str_with_body()


class OpenApiFormatError(OssError):
    def __init__(self, message):
        OssError.__init__(self, OSS_FORMAT_ERROR_STATUS, {}, message, {})

    def __str__(self):
        return self._str_with_body()


class OpenApiServerError(OssError):
    def __init__(self, status, request_id, message, error_code):
        OssError.__init__(self, status, {OSS_REQUEST_ID : request_id}, '', {'Code': error_code, 'Message': message})


class ServerError(OssError):
    pass


class NotFound(ServerError):
    status = 404
    code = ''


class MalformedXml(ServerError):
    status = 400
    code = 'MalformedXML'


class InvalidRequest(ServerError):
    status = 400
    code = 'InvalidRequest'


class OperationNotSupported(ServerError):
    status = 400
    code = 'OperationNotSupported'


class RestoreAlreadyInProgress(ServerError):
    status = 409
    code = 'RestoreAlreadyInProgress'


class InvalidArgument(ServerError):
    status = 400
    code = 'InvalidArgument'

    def __init__(self, status, headers, body, details):
        super(InvalidArgument, self).__init__(status, headers, body, details)
        self.name = details.get('ArgumentName')
        self.value = details.get('ArgumentValue')


class InvalidDigest(ServerError):
    status = 400
    code = 'InvalidDigest'


class InvalidObjectName(ServerError):
    status = 400
    code = 'InvalidObjectName'


class NotImplemented(ServerError):
    status = 400
    code = 'NotImplemented'


class InvalidEncryptionRequest(ServerError):
    status = 400
    code = 'InvalidEncryptionRequest'

class BucketReplicationAlreadyExist(ServerError):
    status = 400
    code = 'BucketReplicationAlreadyExist'

class NoSuchBucket(NotFound):
    status = 404
    code = 'NoSuchBucket'


class NoSuchKey(NotFound):
    status = 404
    code = 'NoSuchKey'


class NoSuchUpload(NotFound):
    status = 404
    code = 'NoSuchUpload'


class NoSuchWebsite(NotFound):
    status = 404
    code = 'NoSuchWebsiteConfiguration'


class NoSuchLifecycle(NotFound):
    status = 404
    code = 'NoSuchLifecycle'


class NoSuchCors(NotFound):
    status = 404
    code = 'NoSuchCORSConfiguration'


class NoSuchLiveChannel(NotFound):
    status = 404
    code = 'NoSuchLiveChannel'


class NoSuchBucketPolicy(NotFound):
    status = 404
    code = 'NoSuchBucketPolicy'

class NoSuchInventory(NotFound):
    status = 404
    code = 'NoSuchInventory'

class NoSuchReplicationRule(NotFound):
    status = 404
    code = 'NoSuchReplicationRule'

class Conflict(ServerError):
    status = 409
    code = ''


class BucketNotEmpty(Conflict):
    status = 409
    code = 'BucketNotEmpty'


class PositionNotEqualToLength(Conflict):
    status = 409
    code = 'PositionNotEqualToLength'

    def __init__(self, status, headers, body, details):
        super(PositionNotEqualToLength, self).__init__(status, headers, body, details)
        self.next_position = int(headers[OSS_NEXT_APPEND_POSITION])


class ObjectNotAppendable(Conflict):
    status = 409
    code = 'ObjectNotAppendable'


class ChannelStillLive(Conflict):
    status = 409
    code = 'ChannelStillLive'


class LiveChannelDisabled(Conflict):
    status = 409
    code = 'LiveChannelDisabled'


class PreconditionFailed(ServerError):
    status = 412
    code = 'PreconditionFailed'


class NotModified(ServerError):
    status = 304
    code = ''


class AccessDenied(ServerError):
    status = 403
    code = 'AccessDenied'

class NoSuchServerSideEncryptionRule(NotFound):
    status = 404
    code = 'NoSuchServerSideEncryptionRule'

class InvalidEncryptionAlgorithmError(ServerError):
    status = 400
    code = 'InvalidEncryptionAlgorithmError' 

class SelectOperationFailed(ServerError):
    code = 'SelectOperationFailed'
    def __init__(self, status, code, message):
        self.status = status
        self.code = code
        self.message = message

    def __str__(self):
        error = {'status': self.status,
                 'code': self.code,
                 'details': self.message}
        return str(error)

class SelectOperationClientError(OssError):
    def __init__(self, message, request_id):
        OssError.__init__(self, OSS_SELECT_CLIENT_ERROR_STATUS, {'x-oss-request-id': request_id}, 'SelectOperationClientError: ' + message, {})
        
    def __str__(self):
        error = {'x-oss-request-id':self.request_id,
                'message': self.message}
        return str(error)

class SignatureDoesNotMatch(ServerError):
    status = 403
    code = 'SignatureDoesNotMatch'

class ObjectAlreadyExists(ServerError):
    status = 400
    code = 'ObjectAlreadyExists'

class PartNotSequential(ServerError):
    status = 400
    code = 'PartNotSequential'

class NoSuchWORMConfiguration(ServerError):
    status = 404
    code = 'NoSuchWORMConfiguration'

class WORMConfigurationLocked(ServerError):
    status = 403
    code = 'WORMConfigurationLocked'

class InvalidWORMConfiguration(ServerError):
    status = 400
    code = 'InvalidWORMConfiguration'

class NoSuchTransferAccelerationConfiguration(ServerError):
    status = 404
    code = 'NoSuchTransferAccelerationConfiguration'

def make_exception(resp):
    status = resp.status
    headers = resp.headers
    body = resp.read(4096)
    details = _parse_error_body(body)
    code = details.get('Code', '')

    try:
        klass = _OSS_ERROR_TO_EXCEPTION[(status, code)]
        return klass(status, headers, body, details)
    except KeyError:
        return ServerError(status, headers, body, details)


def _walk_subclasses(klass):
    for sub in klass.__subclasses__():
        yield sub
        for subsub in _walk_subclasses(sub):
            yield subsub


for klass in _walk_subclasses(ServerError):
    status = getattr(klass, 'status', None)
    code = getattr(klass, 'code', None)

    if status is not None and code is not None:
        _OSS_ERROR_TO_EXCEPTION[(status, code)] = klass


# XML parsing exceptions have changed in Python2.7 and ElementTree 1.3
if hasattr(ElementTree, 'ParseError'):
    ElementTreeParseError = (ElementTree.ParseError, expat.ExpatError)
else:
    ElementTreeParseError = (expat.ExpatError)


def _parse_error_body(body):
    try:
        root = ElementTree.fromstring(body)
        if root.tag != 'Error':
            return {}

        details = {}
        for child in root:
            details[child.tag] = child.text
        return details
    except ElementTreeParseError:
        return _guess_error_details(body)


def _guess_error_details(body):
    details = {}
    body = to_string(body)

    if '<Error>' not in body or '</Error>' not in body:
        return details

    m = re.search('<Code>(.*)</Code>', body)
    if m:
        details['Code'] = m.group(1)

    m = re.search('<Message>(.*)</Message>', body)
    if m:
        details['Message'] = m.group(1)

    return details



# -*- coding: utf-8 -*-

import unittest
import oss2
from oss2.utils import *


class TestUtils(unittest.TestCase):
    def test_aes_ctr_is_block_aligned(self):
        cipher = AESCTRCipher()
        start = 1024
        is_aligned = cipher.is_block_aligned(start)
        self.assertTrue(is_aligned)

        start = 1025
        is_aligned = cipher.is_block_aligned(start)
        self.assertFalse(is_aligned)
        
    def test_aes_ctr_calc_offset(self):
        cipher = AESCTRCipher()
        start = 1024
        cout_offset = cipher.calc_offset(start)
        self.assertEqual(cout_offset, start / 16)

        start = 1025
        self.assertRaises(ClientError, cipher.calc_offset, start)

    def test_acs_ctr_is_valid_part_size(self):
        cipher = AESCTRCipher()
        self.assertFalse(cipher.is_valid_part_size(1, 1024*1024*100))
        self.assertFalse(cipher.is_valid_part_size(1024 * 100 + 1, 1024*1024*100))
        self.assertFalse(cipher.is_valid_part_size(1024 * 100, 1024*1024*1024))
        self.assertTrue(cipher.is_valid_part_size(1024 * 100, 1024*1024*100))

    def test_acs_ctr_determine_part_size(self):
        cipher = AESCTRCipher()
        self.assertEqual(cipher.determine_part_size(1024*100*100000), 1024*1024*10)
        self.assertEqual(cipher.determine_part_size(1024*100*100000 - 1), 1024*1024*10)
        self.assertEqual(cipher.determine_part_size(1024*100*99), 1024*1024*10)

        self.assertEqual(cipher.determine_part_size(1024*100*1000, 1024*100), 1024*100)
        self.assertEqual(cipher.determine_part_size(1024*100*1000, 1024*100-1), 1024*100)
        self.assertEqual(cipher.determine_part_size(1024*100*10000, 1024), 1024*1024*10)

        oss2.defaults.part_size = 1024*1024 - 1
        self.assertEqual(cipher.determine_part_size(1024 * 1024 * 1000), 1024 * 1024)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import os, sys
parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, parentdir)

from .common import *
from oss2.exceptions import *


class TestLiveChannel(OssTestCase):
    def tearDown(self):
        self._delete_channels()
        OssTestCase.tearDown(self)
            
    def _get_play_url(self, bucket_name, channel_name, playlist_name):
        return 'http://%s.%s/%s/%s' % (bucket_name, 
                                       OSS_ENDPOINT.replace('http://', '').replace('https://', ''), 
                                       channel_name, 
                                       playlist_name if playlist_name else 'playlist.m3u8')

    def _get_publish_url(self, bucket_name, channel_name):
        return 'rtmp://%s.%s/live/%s' % (bucket_name, 
                                         OSS_ENDPOINT.replace('http://', '').replace('https://', ''),
                                         channel_name)

    def _get_fixed_number(self, size, n):
        nstr = str(n)
        if size > len(nstr):
            nstr = (size - len(nstr)) * '0' + nstr
        return nstr
    
    def _delete_channels(self):
        prefix = ''
        max_keys = 1000
        for info in oss2.LiveChannelIterator(self.bucket, prefix, max_keys=max_keys):
            self.bucket.delete_live_channel(info.name)

    def _assert_list_result(self,
                            result,
                            marker = '',
                            prefix = '',
                            next_marker = '',
                            max_keys = 0,
                            is_truncated = False,
                            return_count = 0):
        self.assertEqual(result.prefix, prefix)
        self.assertEqual(result.marker, marker)
        self.assertEqual(result.next_marker, next_marker)
        self.assertEqual(result.max_keys, max_keys)
        self.assertEqual(result.is_truncated, is_truncated)
        self.assertEqual(len(result.channels), return_count)

    def test_create_live_channel(self):
        channel_name = 'test-create-channel'
        playlist_name = 'test.m3u8'
        
        channel_target = oss2.models.LiveChannelInfoTarget(playlist_name = playlist_name)
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        create_result = self.bucket.create_live_channel(channel_name, channel_info)

        self.assertEqual(create_result.play_url,
                         self._get_play_url(self.bucket.bucket_name, channel_name, playlist_name))
        self.assertEqual(create_result.publish_url,
                         self._get_publish_url(self.bucket.bucket_name, channel_name))

        self.bucket.delete_live_channel(channel_name)

    def test_get_live_channel(self):
        channel_name = 'test-get-channel'

        self.assertRaises(NoSuchLiveChannel, self.bucket.get_live_channel, channel_name)

        channel_target = oss2.models.LiveChannelInfoTarget(playlist_name = 'test.m3u8')
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        self.bucket.create_live_channel(channel_name, channel_info)

        get_result = self.bucket.get_live_channel(channel_name)
        self.assertEqual(get_result.description, channel_info.description)
        self.assertEqual(get_result.status, channel_info.status)
        self.assertEqual(get_result.target.type, channel_target.type)
        self.assertEqual(get_result.target.frag_duration, str(channel_target.frag_duration))
        self.assertEqual(get_result.target.frag_count, str(channel_target.frag_count))
        self.assertEqual(get_result.target.playlist_name, channel_target.playlist_name)

        self.bucket.delete_live_channel(channel_name)

    def test_list_live_channel(self):
        auth = oss2.Auth(OSS_ID, OSS_SECRET)
        bucket_name = OSS_BUCKET + "-test-list-live-channel"
        bucket = oss2.Bucket(auth, OSS_ENDPOINT, bucket_name)
        bucket.create_bucket(oss2.BUCKET_ACL_PRIVATE)

        channel_name = 'test-list-channel'
        self.assertRaises(NoSuchLiveChannel, bucket.get_live_channel, channel_name)

        list_result = bucket.list_live_channel()
        self._assert_list_result(list_result,
                prefix = '',
                marker = '',
                next_marker = '',
                max_keys = 100,
                is_truncated = False,
                return_count = 0)

        channel_name_list = []
        prefix1 = random_string(5)
        channel_target = oss2.models.LiveChannelInfoTarget(playlist_name = 'test.m3u8')
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        for index in range(200):
            channel_name_list.append(prefix1 + self._get_fixed_number(10, index))
            bucket.create_live_channel(channel_name_list[index], channel_info)

        list_result = bucket.list_live_channel()
        next_marker = prefix1 + self._get_fixed_number(10, 99)
        self._assert_list_result(list_result,
                prefix = '',
                marker = '',
                next_marker = next_marker,
                max_keys = 100,
                is_truncated = True,
                return_count = 100)

        prefix2 = random_string(5)
        list_result = bucket.list_live_channel(prefix = prefix2)
        self._assert_list_result(list_result,
                prefix = prefix2,
                marker = '',
                next_marker = '',
                max_keys = 100,
                is_truncated = False,
                return_count = 0)

        marker = prefix1 + self._get_fixed_number(10, 100)
        list_result = bucket.list_live_channel(
                prefix = prefix1,
                marker = marker)
        self._assert_list_result(list_result,
                prefix = prefix1,
                marker = marker,
                next_marker = '',
                max_keys = 100,
                is_truncated = False,
                return_count = 99)

        max_keys = 1000
        list_result = bucket.list_live_channel(max_keys = max_keys)
        self._assert_list_result(list_result,
                prefix = '',
                marker = '',
                next_marker = '',
                max_keys = max_keys,
                is_truncated = False,
                return_count = 200)

        for channel_name in channel_name_list:
            bucket.delete_live_channel(channel_name)
        bucket.delete_bucket()
        wait_meta_sync()
        self.assertRaises(oss2.exceptions.NoSuchBucket, bucket.delete_bucket)

    def test_get_live_channel_stat(self):
        channel_name = 'test-get-channel-stat'
        playlist_name = 'test.m3u8'
        
        channel_target = oss2.models.LiveChannelInfoTarget(playlist_name = playlist_name)
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        self.bucket.create_live_channel(channel_name, channel_info)

        get_stat_result = self.bucket.get_live_channel_stat(channel_name)
        self.assertEqual(get_stat_result.status, 'Idle')

        self.bucket.delete_live_channel(channel_name)

    def test_put_live_channel_status(self):
        channel_name = 'test-put-channel-status'
        
        channel_target = oss2.models.LiveChannelInfoTarget()
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        self.bucket.create_live_channel(channel_name, channel_info)

        get_result = self.bucket.get_live_channel(channel_name)
        self.assertEqual(get_result.status, 'enabled')

        self.bucket.put_live_channel_status(channel_name, 'disabled')

        get_result = self.bucket.get_live_channel(channel_name)
        self.assertEqual(get_result.status, 'disabled')
        
        self.bucket.put_live_channel_status(channel_name, 'enabled')
        
        get_result = self.bucket.get_live_channel(channel_name)
        self.assertEqual(get_result.status, 'enabled')

        self.bucket.delete_live_channel(channel_name)

    def test_get_live_channel_history(self):
        channel_name = 'test-get-channel-history'
        
        channel_target = oss2.models.LiveChannelInfoTarget()
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        self.bucket.create_live_channel(channel_name, channel_info)

        get_result = self.bucket.get_live_channel_history(channel_name)
        self.assertEqual(len(get_result.records), 0)

        self.bucket.delete_live_channel(channel_name)

    def test_post_vod_playlist(self):
        channel_name = 'test-post-vod-playlist'
        
        channel_target = oss2.models.LiveChannelInfoTarget()
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        self.bucket.create_live_channel(channel_name, channel_info)

        # push rtmp stream here, then generate some ts files on oss.

        end_time = int(time.time()) - 60
        start_time = end_time - 3600
        playlist_name = 'vod_playlist.m3u8'

        # throw exception because no ts file been generated.
        self.assertRaises(oss2.exceptions.InvalidArgument,
                          self.bucket.post_vod_playlist,
                          channel_name,
                          playlist_name,
                          start_time = start_time,
                          end_time = end_time)

        self.bucket.delete_live_channel(channel_name)

    def test_get_vod_playlist(self):
        channel_name = 'test-post-vod-playlist'
        
        channel_target = oss2.models.LiveChannelInfoTarget()
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        self.bucket.create_live_channel(channel_name, channel_info)

        # push rtmp stream here, then generate some ts files on oss.

        end_time = int(time.time()) - 60
        start_time = end_time - 3600

        # throw exception because no ts file been generated.
        self.assertRaises(oss2.exceptions.InvalidArgument,
                          self.bucket.get_vod_playlist,
                          channel_name,
                          start_time = start_time,
                          end_time = end_time)

        self.bucket.delete_live_channel(channel_name)

    def test_sign_rtmp_url(self):
        channel_name = 'test-sign-rtmp-url'
        playlist_name = 'test.m3u8'
        
        channel_target = oss2.models.LiveChannelInfoTarget(playlist_name = playlist_name)
        channel_info = oss2.models.LiveChannelInfo(target = channel_target)
        self.bucket.create_live_channel(channel_name, channel_info)
        
        expires = 3600
        signed_url = self.bucket.sign_rtmp_url(channel_name, playlist_name, expires)
        self.assertTrue(signed_url.startswith(self._get_publish_url(self.bucket.bucket_name, channel_name)))

        # empty playlist name
        signed_url = self.bucket.sign_rtmp_url(channel_name, '', expires)
        self.assertTrue(signed_url.startswith(self._get_publish_url(self.bucket.bucket_name, channel_name)))
         
        self.bucket.delete_live_channel(channel_name)
        
    def test_anonymous_auth(self):
        channel_name = 'test-chan-anonymous-auth'
        playlist_name = 'test.m3u8'
        
        bucket = oss2.Bucket(oss2.AnonymousAuth(), OSS_ENDPOINT, self.bucket.bucket_name)
        signed_url = bucket.sign_rtmp_url(channel_name, playlist_name, 0)
        self.assertEqual(signed_url, self._get_publish_url(self.bucket.bucket_name, channel_name) 
                         + "?playlistName=" + playlist_name)
        
        self.assertRaises(oss2.exceptions.AccessDenied, bucket.delete_live_channel, 'test-live-chan')
    

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import os
import sys
import tempfile

import oss2

# 以下代码展示了进度条功能的用法，包括上传进度条和下载进度条。


# 首先初始化AccessKeyId、AccessKeySecret、Endpoint等信息。
# 通过环境变量获取，或者把诸如“<你的AccessKeyId>”替换成真实的AccessKeyId等。
access_key_id = os.getenv('OSS_TEST_ACCESS_KEY_ID', '<你的AccessKeyId>')
access_key_secret = os.getenv('OSS_TEST_ACCESS_KEY_SECRET', '<你的AccessKeySecret>')
bucket_name = os.getenv('OSS_TEST_BUCKET', '<你的Bucket>')
endpoint = os.getenv('OSS_TEST_ENDPOINT', '<你的访问域名>')


# 确认上面的参数都填写正确了
for param in (access_key_id, access_key_secret, bucket_name, endpoint):
    assert '<' not in param, '请设置参数：' + param

def percentage(consumed_bytes, total_bytes):
    """进度条回调函数，计算当前完成的百分比
    
    :param consumed_bytes: 已经上传/下载的数据量
    :param total_bytes: 总数据量
    """
    if total_bytes:
        rate = int(100 * (float(consumed_bytes) / float(total_bytes)))
        print('\r{0}% '.format(rate))
        sys.stdout.flush()

def _prepare_temp_file(content):
    """创建临时文件
    :param content: 文件内容
    :return 文件名
    """
    fd, pathname = tempfile.mkstemp(suffix='exam-progress-')
    os.write(fd, content)
    os.close(fd)
    return pathname

key = 'story.txt'
content = 'a' * 1024 * 1024

# 创建Bucket对象，所有Object相关的接口都可以通过Bucket对象来进行
bucket = oss2.Bucket(oss2.Auth(access_key_id, access_key_secret), endpoint, bucket_name)

"""
流式上传
"""
# 带有进度条的覆盖上传
bucket.put_object(key, content, progress_callback=percentage)

# 删除上传的文件
bucket.delete_object(key)

"""
追加上传
"""
# 带有进度条的追加上传，每一次追加一个进度条
# 创建可追加文件，首次偏移（position）设为0
result = bucket.append_object(key, 0, content, progress_callback=percentage)
# 追加一行数据，偏移可以从上次响应中获得
# 当然，也可以通过head_object()获得当前长度作为偏移，只是比较低效
bucket.append_object(key, result.next_position, content, progress_callback=percentage)

# 删除上传的文件
bucket.delete_object(key)

"""
分片上传
"""
# 带有进度条的分片上传，每个分片上传一个进度条
parts = []
upload_id = bucket.init_multipart_upload(key).upload_id

# 上传分片
for i in range(3):
    result = bucket.upload_part(key, upload_id, i+1, content, progress_callback=percentage)
    parts.append(oss2.models.PartInfo(i+1, result.etag, size = len(content), part_crc = result.crc))

# 完成上传并回调
result = bucket.complete_multipart_upload(key, upload_id, parts)

"""
断点续传上传
"""
# 带进度条的断点续传
pathname = _prepare_temp_file(content)
oss2.resumable_upload(bucket, key, pathname, 
                      multipart_threshold=200*1024,
                      part_size=100*1024,
                      num_threads=3,
                      progress_callback=percentage)

"""
文件下载
"""
# 带进度条的下载
result = bucket.get_object(key, progress_callback=percentage)
content_got = b''
for chunk in result:
    content_got += chunk
assert content == content_got

"""
范围下载
"""
# 带进度条的范围下载
result = bucket.get_object(key, byte_range=(1024, 2047), progress_callback=percentage)
content_got = b''
for chunk in result:
    content_got += chunk
assert 'a'*1024 == content_got

"""
断点续传下载 
"""
# 带进度条的断点续传下载 
filename = 'download.txt'
oss2.resumable_download(bucket, key, filename,
                        multiget_threshold=200*1024,
                        part_size=100*1024,
                        num_threads=3,
                        progress_callback=percentage) 
os.remove(filename)

# 删除上传的文件
bucket.delete_object(key)



# -*- coding: utf-8 -*-

import hmac
import hashlib
import time

from . import utils
from .compat import urlquote, to_bytes, is_py2
from .headers import *
import logging
from .credentials import StaticCredentialsProvider

AUTH_VERSION_1 = 'v1'
AUTH_VERSION_2 = 'v2'

logger = logging.getLogger(__name__)


def make_auth(access_key_id, access_key_secret, auth_version=AUTH_VERSION_1):
    if auth_version == AUTH_VERSION_2:
        logger.debug("Init Auth V2: access_key_id: {0}, access_key_secret: ******".format(access_key_id))
        return AuthV2(access_key_id.strip(), access_key_secret.strip())
    else:
        logger.debug("Init Auth v1: access_key_id: {0}, access_key_secret: ******".format(access_key_id))
        return Auth(access_key_id.strip(), access_key_secret.strip())


class AuthBase(object):
    """用于保存用户AccessKeyId、AccessKeySecret，以及计算签名的对象。"""
    def __init__(self, credentials_provider):
        self.credentials_provider = credentials_provider

    def _sign_rtmp_url(self, url, bucket_name, channel_name, expires, params):
        credentials = self.credentials_provider.get_credentials()
        if credentials.get_security_token():
            params['security-token'] = credentials.get_security_token()

        expiration_time = int(time.time()) + expires

        canonicalized_resource = "/%s/%s" % (bucket_name, channel_name)
        canonicalized_params = []

        if params:
            items = params.items()
            for k, v in items:
                if k != "OSSAccessKeyId" and k != "Signature" and k != "Expires" and k != "SecurityToken":
                    canonicalized_params.append((k, v))

        canonicalized_params.sort(key=lambda e: e[0])
        canon_params_str = ''
        for k, v in canonicalized_params:
            canon_params_str += '%s:%s\n' % (k, v)

        p = params if params else {}
        string_to_sign = str(expiration_time) + "\n" + canon_params_str + canonicalized_resource
        logger.debug('Sign Rtmp url: string to be signed = {0}'.format(string_to_sign))


        h = hmac.new(to_bytes(credentials.get_access_key_secret()), to_bytes(string_to_sign), hashlib.sha1)
        signature = utils.b64encode_as_string(h.digest())

        p['OSSAccessKeyId'] = credentials.get_access_key_id()
        p['Expires'] = str(expiration_time)
        p['Signature'] = signature

        return url + '?' + '&'.join(_param_to_quoted_query(k, v) for k, v in p.items())



class ProviderAuth(AuthBase):
    """签名版本1
    默认构造函数同父类AuthBase，需要传递credentials_provider
    """
    _subresource_key_set = frozenset(
        ['response-content-type', 'response-content-language',
         'response-cache-control', 'logging', 'response-content-encoding',
         'acl', 'uploadId', 'uploads', 'partNumber', 'group', 'link',
         'delete', 'website', 'location', 'objectInfo', 'objectMeta',
         'response-expires', 'response-content-disposition', 'cors', 'lifecycle',
         'restore', 'qos', 'referer', 'stat', 'bucketInfo', 'append', 'position', 'security-token',
         'live', 'comp', 'status', 'vod', 'startTime', 'endTime', 'x-oss-process',
         'symlink', 'callback', 'callback-var', 'tagging', 'encryption', 'versions',
         'versioning', 'versionId', 'policy', 'requestPayment', 'x-oss-traffic-limit', 'qosInfo', 'asyncFetch',
         'x-oss-request-payer', 'sequential', 'inventory', 'inventoryId', 'continuation-token', 'callback',
         'callback-var', 'worm', 'wormId', 'wormExtend', 'replication', 'replicationLocation',
         'replicationProgress', 'transferAcceleration']
    )

    def _sign_request(self, req, bucket_name, key):
        credentials = self.credentials_provider.get_credentials()
        if credentials.get_security_token():
            req.headers[OSS_SECURITY_TOKEN] = credentials.get_security_token()

        req.headers['date'] = utils.http_date()

        signature = self.__make_signature(req, bucket_name, key, credentials)
        req.headers['authorization'] = "OSS {0}:{1}".format(credentials.get_access_key_id(), signature)

    def _sign_url(self, req, bucket_name, key, expires):
        credentials = self.credentials_provider.get_credentials()
        if credentials.get_security_token():
            req.params['security-token'] = credentials.get_security_token()

        expiration_time = int(time.time()) + expires

        req.headers['date'] = str(expiration_time)
        signature = self.__make_signature(req, bucket_name, key, credentials)

        req.params['OSSAccessKeyId'] = credentials.get_access_key_id()
        req.params['Expires'] = str(expiration_time)
        req.params['Signature'] = signature

        return req.url + '?' + '&'.join(_param_to_quoted_query(k, v) for k, v in req.params.items())

    def __make_signature(self, req, bucket_name, key, credentials):
        if is_py2:
            string_to_sign = self.__get_string_to_sign(req, bucket_name, key)
        else:
            string_to_sign = self.__get_bytes_to_sign(req, bucket_name, key)

        logger.debug('Make signature: string to be signed = {0}'.format(string_to_sign))

        h = hmac.new(to_bytes(credentials.get_access_key_secret()), to_bytes(string_to_sign), hashlib.sha1)
        return utils.b64encode_as_string(h.digest())

    def __get_string_to_sign(self, req, bucket_name, key):
        resource_string = self.__get_resource_string(req, bucket_name, key)
        headers_string = self.__get_headers_string(req)

        content_md5 = req.headers.get('content-md5', '')
        content_type = req.headers.get('content-type', '')
        date = req.headers.get('date', '')
        return '\n'.join([req.method,
                          content_md5,
                          content_type,
                          date,
                          headers_string + resource_string])

    def __get_headers_string(self, req):
        headers = req.headers
        canon_headers = []
        for k, v in headers.items():
            lower_key = k.lower()
            if lower_key.startswith('x-oss-'):
                canon_headers.append((lower_key, v))

        canon_headers.sort(key=lambda x: x[0])

        if canon_headers:
            return '\n'.join(k + ':' + v for k, v in canon_headers) + '\n'
        else:
            return ''

    def __get_resource_string(self, req, bucket_name, key):
        if not bucket_name:
            return '/' + self.__get_subresource_string(req.params)
        else:
            return '/{0}/{1}{2}'.format(bucket_name, key, self.__get_subresource_string(req.params))

    def __get_subresource_string(self, params):
        if not params:
            return ''

        subresource_params = []
        for key, value in params.items():
            if key in self._subresource_key_set:
                subresource_params.append((key, value))

        subresource_params.sort(key=lambda e: e[0])

        if subresource_params:
            return '?' + '&'.join(self.__param_to_query(k, v) for k, v in subresource_params)
        else:
            return ''

    def __param_to_query(self, k, v):
        if v:
            return k + '=' + v
        else:
            return k

    def __get_bytes_to_sign(self, req, bucket_name, key):
        resource_bytes = self.__get_resource_string(req, bucket_name, key).encode('utf-8')
        headers_bytes = self.__get_headers_bytes(req)

        content_md5 = req.headers.get('content-md5', '').encode('utf-8')
        content_type = req.headers.get('content-type', '').encode('utf-8')
        date = req.headers.get('date', '').encode('utf-8')
        return b'\n'.join([req.method.encode('utf-8'),
                          content_md5,
                          content_type,
                          date,
                          headers_bytes + resource_bytes])

    def __get_headers_bytes(self, req):
        headers = req.headers
        canon_headers = []
        for k, v in headers.items():
            lower_key = k.lower()
            if lower_key.startswith('x-oss-'):
                canon_headers.append((lower_key, v))

        canon_headers.sort(key=lambda x: x[0])

        if canon_headers:
            return b'\n'.join(to_bytes(k) + b':' + to_bytes(v) for k, v in canon_headers) + b'\n'
        else:
            return b''

class Auth(ProviderAuth):
    """签名版本1
    """
    def __init__(self, access_key_id, access_key_secret):
        credentials_provider = StaticCredentialsProvider(access_key_id.strip(), access_key_secret.strip())
        super(Auth, self).__init__(credentials_provider)


class AnonymousAuth(object):
    """用于匿名访问。

    .. note::
        匿名用户只能读取public-read的Bucket，或只能读取、写入public-read-write的Bucket。
        不能进行Service、Bucket相关的操作，也不能罗列文件等。
    """
    def _sign_request(self, req, bucket_name, key):
        pass

    def _sign_url(self, req, bucket_name, key, expires):
        return req.url + '?' + '&'.join(_param_to_quoted_query(k, v) for k, v in req.params.items())
    
    def _sign_rtmp_url(self, url, bucket_name, channel_name, expires, params):
        return url + '?' + '&'.join(_param_to_quoted_query(k, v) for k, v in params.items())


class StsAuth(object):
    """用于STS临时凭证访问。可以通过官方STS客户端获得临时密钥（AccessKeyId、AccessKeySecret）以及临时安全令牌（SecurityToken）。

    注意到临时凭证会在一段时间后过期，在此之前需要重新获取临时凭证，并更新 :class:`Bucket <oss2.Bucket>` 的 `auth` 成员变量为新
    的 `StsAuth` 实例。

    :param str access_key_id: 临时AccessKeyId
    :param str access_key_secret: 临时AccessKeySecret
    :param str security_token: 临时安全令牌(SecurityToken)
    :param str auth_version: 需要生成auth的版本，默认为AUTH_VERSION_1(v1)
    """
    def __init__(self, access_key_id, access_key_secret, security_token, auth_version=AUTH_VERSION_1):
        logger.debug("Init StsAuth: access_key_id: {0}, access_key_secret: ******, security_token: ******".format(access_key_id))
        credentials_provider = StaticCredentialsProvider(access_key_id, access_key_secret, security_token)
        self.__auth = ProviderAuthV2(credentials_provider) if auth_version == AUTH_VERSION_2 else ProviderAuth(credentials_provider)

    def _sign_request(self, req, bucket_name, key):
        self.__auth._sign_request(req, bucket_name, key)

    def _sign_url(self, req, bucket_name, key, expires):
        return self.__auth._sign_url(req, bucket_name, key, expires)

    def _sign_rtmp_url(self, url, bucket_name, channel_name, expires, params):
        return self.__auth._sign_rtmp_url(url, bucket_name, channel_name, expires, params)


def _param_to_quoted_query(k, v):
    if v:
        return urlquote(k, '') + '=' + urlquote(v, '')
    else:
        return urlquote(k, '')


def v2_uri_encode(raw_text):
    raw_text = to_bytes(raw_text)

    res = ''
    for b in raw_text:
        if isinstance(b, int):
            c = chr(b)
        else:
            c = b

        if (c >= 'A' and c <= 'Z') or (c >= 'a' and c <= 'z')\
            or (c >= '0' and c <= '9') or c in ['_', '-', '~', '.']:
            res += c
        else:
            res += "%{0:02X}".format(ord(c))

    return res


_DEFAULT_ADDITIONAL_HEADERS = set(['range',
                                   'if-modified-since'])


class ProviderAuthV2(AuthBase):
    """签名版本2，默认构造函数同父类AuthBase，需要传递credentials_provider
    与版本1的区别在：
    1. 使用SHA256算法，具有更高的安全性
    2. 参数计算包含所有的HTTP查询参数
    """
    def _sign_request(self, req, bucket_name, key, in_additional_headers=None):
        """把authorization放入req的header里面

        :param req: authorization信息将会加入到这个请求的header里面
        :type req: oss2.http.Request

        :param bucket_name: bucket名称
        :param key: OSS文件名
        :param in_additional_headers: 加入签名计算的额外header列表
        """
        credentials = self.credentials_provider.get_credentials()
        if credentials.get_security_token():
            req.headers[OSS_SECURITY_TOKEN] = credentials.get_security_token()

        if in_additional_headers is None:
            in_additional_headers = _DEFAULT_ADDITIONAL_HEADERS

        additional_headers = self.__get_additional_headers(req, in_additional_headers)

        req.headers['date'] = utils.http_date()

        signature = self.__make_signature(req, bucket_name, key, additional_headers, credentials)

        if additional_headers:
            req.headers['authorization'] = "OSS2 AccessKeyId:{0},AdditionalHeaders:{1},Signature:{2}"\
                .format(credentials.get_access_key_id(), ';'.join(additional_headers), signature)
        else:
            req.headers['authorization'] = "OSS2 AccessKeyId:{0},Signature:{1}".format(credentials.get_access_key_id(), signature)

    def _sign_url(self, req, bucket_name, key, expires, in_additional_headers=None):
        """返回一个签过名的URL

        :param req: 需要签名的请求
        :type req: oss2.http.Request

        :param bucket_name: bucket名称
        :param key: OSS文件名
        :param int expires: 返回的url将在`expires`秒后过期.
        :param in_additional_headers: 加入签名计算的额外header列表

        :return: a signed URL
        """
        credentials = self.credentials_provider.get_credentials()
        if credentials.get_security_token():
            req.params['security-token'] = credentials.get_security_token()

        if in_additional_headers is None:
            in_additional_headers = set()

        additional_headers = self.__get_additional_headers(req, in_additional_headers)

        expiration_time = int(time.time()) + expires

        req.headers['date'] = str(expiration_time)  # re-use __make_signature by setting the 'date' header

        req.params['x-oss-signature-version'] = 'OSS2'
        req.params['x-oss-expires'] = str(expiration_time)
        req.params['x-oss-access-key-id'] = credentials.get_access_key_id()

        signature = self.__make_signature(req, bucket_name, key, additional_headers, credentials)

        req.params['x-oss-signature'] = signature

        return req.url + '?' + '&'.join(_param_to_quoted_query(k, v) for k, v in req.params.items())

    def __make_signature(self, req, bucket_name, key, additional_headers, credentials):
        if is_py2:
            string_to_sign = self.__get_string_to_sign(req, bucket_name, key, additional_headers)
        else:
            string_to_sign = self.__get_bytes_to_sign(req, bucket_name, key, additional_headers)

        logger.debug('Make signature: string to be signed = {0}'.format(string_to_sign))

        h = hmac.new(to_bytes(credentials.get_access_key_secret()), to_bytes(string_to_sign), hashlib.sha256)
        return utils.b64encode_as_string(h.digest())

    def __get_additional_headers(self, req, in_additional_headers):
        # we add a header into additional_headers only if it is already in req's headers.

        additional_headers = set(h.lower() for h in in_additional_headers)
        keys_in_header = set(k.lower() for k in req.headers.keys())

        return additional_headers & keys_in_header

    def __get_string_to_sign(self, req, bucket_name, key, additional_header_list):
        verb = req.method
        content_md5 = req.headers.get('content-md5', '')
        content_type = req.headers.get('content-type', '')
        date = req.headers.get('date', '')

        canonicalized_oss_headers = self.__get_canonicalized_oss_headers(req, additional_header_list)
        additional_headers = ';'.join(sorted(additional_header_list))
        canonicalized_resource = self.__get_resource_string(req, bucket_name, key)

        return verb + '\n' +\
            content_md5 + '\n' +\
            content_type + '\n' +\
            date + '\n' +\
            canonicalized_oss_headers +\
            additional_headers + '\n' +\
            canonicalized_resource

    def __get_resource_string(self, req, bucket_name, key):
        if bucket_name:
            encoded_uri = v2_uri_encode('/' + bucket_name + '/' + key)
        else:
            encoded_uri = v2_uri_encode('/')

        logger.info('encoded_uri={0} key={1}'.format(encoded_uri, key))

        return encoded_uri + self.__get_canonalized_query_string(req)

    def __get_canonalized_query_string(self, req):
        encoded_params = {}
        for param, value in req.params.items():
            encoded_params[v2_uri_encode(param)] = v2_uri_encode(value)

        if not encoded_params:
            return ''

        sorted_params = sorted(encoded_params.items(), key=lambda e: e[0])
        return '?' + '&'.join(self.__param_to_query(k, v) for k, v in sorted_params)

    def __param_to_query(self, k, v):
        if v:
            return k + '=' + v
        else:
            return k

    def __get_canonicalized_oss_headers(self, req, additional_headers):
        """
        :param additional_headers: 小写的headers列表, 并且这些headers都不以'x-oss-'为前缀.
        """
        canon_headers = []

        for k, v in req.headers.items():
            lower_key = k.lower()
            if lower_key.startswith('x-oss-') or lower_key in additional_headers:
                canon_headers.append((lower_key, v))

        canon_headers.sort(key=lambda x: x[0])

        return ''.join(v[0] + ':' + v[1] + '\n' for v in canon_headers)

    def __get_bytes_to_sign(self, req, bucket_name, key, additional_header_list):
        verb = req.method.encode('utf-8')
        content_md5 = req.headers.get('content-md5', '').encode('utf-8')
        content_type = req.headers.get('content-type', '').encode('utf-8')
        date = req.headers.get('date', '').encode('utf-8')

        canonicalized_oss_headers = self.__get_canonicalized_oss_headers_bytes(req, additional_header_list)
        additional_headers = ';'.join(sorted(additional_header_list)).encode('utf-8')
        canonicalized_resource = self.__get_resource_string(req, bucket_name, key).encode('utf-8')

        return verb + b'\n' +\
            content_md5 + b'\n' +\
            content_type + b'\n' +\
            date + b'\n' +\
            canonicalized_oss_headers +\
            additional_headers + b'\n' +\
            canonicalized_resource

    def __get_canonicalized_oss_headers_bytes(self, req, additional_headers):
        """
        :param additional_headers: 小写的headers列表, 并且这些headers都不以'x-oss-'为前缀.
        """
        canon_headers = []

        for k, v in req.headers.items():
            lower_key = k.lower()
            if lower_key.startswith('x-oss-') or lower_key in additional_headers:
                canon_headers.append((lower_key, v))

        canon_headers.sort(key=lambda x: x[0])

        return b''.join(to_bytes(v[0]) + b':' + to_bytes(v[1]) + b'\n' for v in canon_headers)


class AuthV2(ProviderAuthV2):
    """签名版本2，与版本1的区别在：
    1. 使用SHA256算法，具有更高的安全性
    2. 参数计算包含所有的HTTP查询参数
    """
    def __init__(self, access_key_id, access_key_secret):
        credentials_provider = StaticCredentialsProvider(access_key_id.strip(), access_key_secret.strip())
        super(AuthV2, self).__init__(credentials_provider)



# -*- coding: utf-8 -*-

import threading
import sys
import logging

logger = logging.getLogger(__name__)

try:
    import Queue as queue
except ImportError:
    import queue

import traceback


class TaskQueue(object):
    def __init__(self, producer, consumers):
        self.__producer = producer
        self.__consumers = consumers

        self.__threads = []

        # must be an infinite queue, otherwise producer may be blocked after all consumers being dead.
        self.__queue = queue.Queue()

        self.__lock = threading.Lock()
        self.__exc_info = None
        self.__exc_stack = ''

    def run(self):
        self.__add_and_run(threading.Thread(target=self.__producer_func))

        for c in self.__consumers:
            self.__add_and_run(threading.Thread(target=self.__consumer_func, args=(c,)))

        # give KeyboardInterrupt chances to happen by joining with timeouts.
        while self.__any_active():
            for t in self.__threads:
                t.join(1)

        if self.__exc_info:
            logger.error('An exception was thrown by producer or consumer, backtrace: {0}'.format(self.__exc_stack))
            raise self.__exc_info[1]

    def put(self, data):
        assert data is not None
        self.__queue.put(data)

    def get(self):
        return self.__queue.get()

    def ok(self):
        with self.__lock:
            return self.__exc_info is None

    def __add_and_run(self, thread):
        thread.daemon = True
        thread.start()
        self.__threads.append(thread)

    def __any_active(self):
        return any(t.is_alive() for t in self.__threads)

    def __producer_func(self):
        try:
            self.__producer(self)
        except:
            self.__on_exception(sys.exc_info())
            self.__put_end()
        else:
            self.__put_end()

    def __consumer_func(self, consumer):
        try:
            consumer(self)
        except:
            self.__on_exception(sys.exc_info())

    def __put_end(self):
        for i in range(len(self.__consumers)):
            self.__queue.put(None)

    def __on_exception(self, exc_info):
        with self.__lock:
            if self.__exc_info is None:
                self.__exc_info = exc_info
                self.__exc_stack = traceback.format_exc()






# -*- coding: utf-8 -*-
class SelectParameters(object):  # Select API中参数名常量，参数名是大小写敏感的。具体含义参加api.py中的介绍
    CsvHeaderInfo = 'CsvHeaderInfo'
    CommentCharacter = 'CommentCharacter'
    RecordDelimiter = 'RecordDelimiter'
    OutputRecordDelimiter = 'OutputRecordDelimiter'
    FieldDelimiter = 'FieldDelimiter'
    OutputFieldDelimiter = 'OutputFieldDelimiter'
    QuoteCharacter = 'QuoteCharacter'
    SplitRange = 'SplitRange' 
    LineRange = 'LineRange'
    CompressionType = 'CompressionType'
    KeepAllColumns = 'KeepAllColumns'
    OutputRawData = 'OutputRawData'
    EnablePayloadCrc = 'EnablePayloadCrc'
    OutputHeader = 'OutputHeader'
    SkipPartialDataRecord = 'SkipPartialDataRecord'
    MaxSkippedRecordsAllowed = 'MaxSkippedRecordsAllowed'
    AllowQuotedRecordDelimiter = 'AllowQuotedRecordDelimiter'
    Json_Type = 'Json_Type'
    ParseJsonNumberAsString = 'ParseJsonNumberAsString'
    OverwriteIfExists = 'OverwriteIfExists'
    AllowQuotedRecordDelimiter = 'AllowQuotedRecordDelimiter'

class SelectJsonTypes(object):  # Select JSOn API中 Json_Type的合法值，大小写敏感。
    DOCUMENT = 'DOCUMENT'
    LINES = 'LINES'



# -*- coding: utf-8 -*-

import os
import oss2
from oss2.models import OSS_TRAFFIC_LIMIT

# 以下代码展示了限速上传下载文件的设置方法

# 首先初始化AccessKeyId、AccessKeySecret、Endpoint等信息。
# 通过环境变量获取，或者把诸如“<你的AccessKeyId>”替换成真实的AccessKeyId等。
access_key_id = os.getenv('OSS_TEST_ACCESS_KEY_ID', '<你的AccessKeyId>')
access_key_secret = os.getenv('OSS_TEST_ACCESS_KEY_SECRET', '<你的AccessKeySecret>')
bucket_name = os.getenv('OSS_TEST_BUCKET', '<你要请求的Bucket名称>')
endpoint = os.getenv('OSS_TEST_ENDPOINT', '<你的访问域名>')

# 确认上面的参数都填写正确了
for param in (access_key_id, access_key_secret, bucket_name, endpoint):
    assert '<' not in param, '请设置参数：' + param

# 创建Bucket对象，所有Object相关的接口都可以通过Bucket对象来进行
bucket = oss2.Bucket(oss2.Auth(access_key_id, access_key_secret), endpoint, bucket_name)

OBJECT_SIZE_1MB = (1 * 1024 * 1024)
LIMIT_100KB = (100 * 1024 * 8)

headers = dict()
headers[OSS_TRAFFIC_LIMIT] = str(LIMIT_100KB);

key = 'traffic-limit-test-put-object'
content = b'a' * OBJECT_SIZE_1MB

# 限速上传文件
result = bucket.put_object(key, content, headers=headers)
print('http response status:', result.status)

# 限速下载文件到本地
file_name = key + '.txt'
result = bucket.get_object_to_file(key, file_name, headers=headers)
print('http response status:', result.status)

os.remove(file_name)
bucket.delete_object(key)

# 使用签名url方式限速上传文件
params = dict()
params[OSS_TRAFFIC_LIMIT] = str(LIMIT_100KB);
local_file_name = "example.jpg"

# 创建限速上传文件的签名url, 有效期60s
url = bucket.sign_url('PUT', key, 60, params=params)
# 限速上传
result = bucket.put_object_with_url_from_file(url, local_file_name)
print('http response status:', result.status)

# 创建限速下载文件的签名url, 有效期60s
down_file_name = key + '.tmp'
url = bucket.sign_url('GET', key, 60, params=params)
# 限速下载
result = bucket.get_object_with_url_to_file(url, down_file_name)
print('http response status:', result.status)

os.remove(down_file_name)
bucket.delete_object(key)


# -*- coding: utf-8 -*-

import oss2
from oss2 import to_string
from .common import *
from oss2.models import (ConditionInlcudeHeader, 
                        Condition, 
                        Redirect, 
                        RedirectMirrorHeaders,
                        MirrorHeadersSet, 
                        RoutingRule, 
                        BucketWebsite,
                        REDIRECT_TYPE_MIRROR,
                        REDIRECT_TYPE_EXTERNAL,
                        REDIRECT_TYPE_ALICDN,
                        REDIRECT_TYPE_INTERNAL)


class TestWebsite(OssTestCase):
    def test_normal_set_website_without_routing_rule(self):
        key = self.random_key('/')
        content = random_bytes(32)

        self.bucket.put_object('index.html', content)

        # set index_file and error_file
        self.bucket.put_bucket_website(BucketWebsite('index.html', 'error.html'))
        wait_meta_sync()

        def same_website(website, index, error):
            return website.index_file == index and website.error_file == error

        # check index_file and error_file
        self.retry_assert(lambda: same_website(self.bucket.get_bucket_website(), 'index.html', 'error.html'))

        # will be reidrect to inext
        result = self.bucket.get_object(key)
        self.assertEqual(result.read(), content)

        self.bucket.delete_object('index.html')

        # test chinese
        for index, error in [('index+中文.html', 'error.中文'), (u'index+中文.html', u'error.中文')]:
            self.bucket.put_bucket_website(BucketWebsite(index, error))
            self.retry_assert(lambda: same_website(self.bucket.get_bucket_website(), to_string(index), to_string(error)))

        # delete website
        self.bucket.delete_bucket_website()
        self.bucket.delete_bucket_website()

        self.assertRaises(oss2.exceptions.NoSuchWebsite, self.bucket.get_bucket_website)

    def test_normal_set_website_with_mirror(self):

        index_file = 'index.html'
        error_file = 'error.html'

        # rule condition
        include_header1= ConditionInlcudeHeader('host', 'test.oss-cn-beijing-internal.aliyuncs.com')
        include_header2 = ConditionInlcudeHeader('host', 'test.oss-cn-shenzhen-internal.aliyuncs.com')
        condition1 = Condition(key_prefix_equals='1~!@#$%^&*()-_=+|\\[]{}<>,./?`~',
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])
        condition2 = Condition(key_prefix_equals='2~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])

        # rule redirect
        mirror_headers_set_1 = MirrorHeadersSet("myheader-key5","myheader-value5")
        mirror_headers_set_2 = MirrorHeadersSet("myheader-key6","myheader-value6")
        set_list = [mirror_headers_set_1, mirror_headers_set_2]
        pass_list = ['myheader-key1', 'myheader-key2']
        remove_list = ['myheader-key3', 'myheader-key4']
        mirror_header = RedirectMirrorHeaders(pass_all=True, pass_list=pass_list, remove_list=remove_list, set_list=set_list)

        # direct type Mirror
        redirect1 = Redirect(redirect_type='Mirror', mirror_url='http://www.test.com/', mirror_pass_query_string=True, 
                            mirror_follow_redirect=True, mirror_check_md5=False)
        
        redirect2 = Redirect(redirect_type='Mirror', pass_query_string=False, mirror_url='http://www.test.com/', 
                            mirror_url_slave='http://www.slave.com/', mirror_url_probe='http://www.test.com/index.html', 
                            mirror_pass_query_string=False, mirror_follow_redirect=True, mirror_check_md5=True, 
                            mirror_headers=mirror_header)

        rule1_num = 1
        rule2_num = 2
        rule1 = RoutingRule(rule1_num, condition=condition1, redirect=redirect1)
        rule2 = RoutingRule(rule2_num, condition=condition2, redirect=redirect2)
        website_set = BucketWebsite(index_file, error_file, [rule1, rule2])
        self.bucket.put_bucket_website(website_set)
        wait_meta_sync()

        # get websete config
        website_get = self.bucket.get_bucket_website()
        self.assertEqual(website_get.index_file, index_file)
        self.assertEqual(website_get.error_file, error_file)
        self.assertEqual(len(website_get.rules), 2)
        
        for rule in website_get.rules:
            # check rule_num
            self.assertTrue(rule.rule_num in [rule1_num, rule2_num])

            if rule.rule_num == rule1_num:
                cmp_condition = condition1
                cmp_redirect = redirect1
            else:
                cmp_condition = condition2
                cmp_redirect = redirect2

            # check conditon
            t_condition = rule.condition
            self.assertEqual(t_condition.key_prefix_equals, cmp_condition.key_prefix_equals)
            self.assertEqual(t_condition.http_err_code_return_equals, cmp_condition.http_err_code_return_equals)
            self.assertEqual(len(t_condition.include_header_list), len(cmp_condition.include_header_list))

            # check redirect
            t_redirect = rule.redirect
            self.assertEqual(t_redirect.redirect_type, cmp_redirect.redirect_type)   
            # pass_query_string will be coved by mirror_pass_query_string when Mirror reidrect type
            self.assertEqual(t_redirect.pass_query_string, cmp_redirect.mirror_pass_query_string)
            self.assertEqual(t_redirect.mirror_url, cmp_redirect.mirror_url)
            self.assertEqual(t_redirect.mirror_url_slave, cmp_redirect.mirror_url_slave)
            self.assertEqual(t_redirect.mirror_url_probe, cmp_redirect.mirror_url_probe)
            self.assertEqual(t_redirect.mirror_pass_query_string, cmp_redirect.mirror_pass_query_string)
            self.assertEqual(t_redirect.mirror_follow_redirect, cmp_redirect.mirror_follow_redirect)   
            self.assertEqual(t_redirect.mirror_check_md5, cmp_redirect.mirror_check_md5)

            #check redirect mirror_headers
            if cmp_redirect.mirror_headers is None:
                continue

            t_mirror_headers = t_redirect.mirror_headers
            cmp_mirror_headers = cmp_redirect.mirror_headers
            self.assertEqual(t_mirror_headers.pass_all, cmp_mirror_headers.pass_all)
            self.assertEqual(len(t_mirror_headers.pass_list), len(cmp_mirror_headers.pass_list))
            self.assertEqual(len(t_mirror_headers.remove_list), len(cmp_mirror_headers.remove_list))
            self.assertEqual(len(t_mirror_headers.set_list), len(cmp_mirror_headers.set_list))

        self.bucket.delete_bucket_website() 

    def test_normal_set_website_with_alicdn_external_internal(self):
        index_file = 'index.html'
        error_file = 'error.html'

        include_header1= ConditionInlcudeHeader('host', 'test.oss-cn-beijing-internal.aliyuncs.com')
        include_header2 = ConditionInlcudeHeader('host', 'test.oss-cn-shenzhen-internal.aliyuncs.com')
        condition1 = Condition(key_prefix_equals='1~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])
        condition2 = Condition(key_prefix_equals='2~!@#$%^&*()-_=+|\\[]{}<>,./?`~',
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])
        condition3 = Condition(key_prefix_equals='3~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])

        # test direct type AliCDN 
        redirect1 = Redirect(redirect_type=REDIRECT_TYPE_ALICDN, pass_query_string=True, replace_key_with='${key}.suffix', 
                            proto='http', host_name='oss.aliyuncs.com', http_redirect_code=302)

        # test direct type  External
        redirect2 = Redirect(redirect_type=REDIRECT_TYPE_EXTERNAL, pass_query_string=False, replace_key_prefix_with='abc',
                            proto='https', host_name='oss.aliyuncs.com', http_redirect_code=302)

        # test direct type  Internal
        redirect3 = Redirect(redirect_type=REDIRECT_TYPE_INTERNAL, pass_query_string=False, replace_key_with='${key}.suffix')

        rule1_num = 1
        rule2_num = 2
        rule3_num = 3
        rule1 = RoutingRule(rule1_num, condition=condition1, redirect=redirect1)
        rule2 = RoutingRule(rule2_num, condition=condition2, redirect=redirect2)
        rule3 = RoutingRule(rule3_num, condition=condition3, redirect=redirect3)
        website_set = BucketWebsite(index_file, error_file, [rule1, rule2, rule3])
        self.bucket.put_bucket_website(website_set)
        wait_meta_sync()

        website_get = self.bucket.get_bucket_website()
        self.assertEqual(website_get.index_file, index_file)
        self.assertEqual(website_get.error_file, error_file)
        self.assertEqual(len(website_get.rules), 3)

        for rule in website_get.rules:
            # check rule_num
            self.assertTrue(rule.rule_num in [rule1_num, rule2_num, rule3_num])

            if rule.rule_num == rule1_num:
                cmp_condition = condition1
                cmp_redirect = redirect1
            elif rule.rule_num == rule2_num:
                cmp_condition = condition2
                cmp_redirect = redirect2
            else:
                cmp_condition = condition3
                cmp_redirect = redirect3

            # check conditon
            t_condition = rule.condition
            self.assertEqual(t_condition.key_prefix_equals, cmp_condition.key_prefix_equals)
            self.assertEqual(t_condition.http_err_code_return_equals, cmp_condition.http_err_code_return_equals)
            self.assertEqual(len(t_condition.include_header_list), len(cmp_condition.include_header_list))

            # check redirect
            t_redirect = rule.redirect
            self.assertEqual(t_redirect.redirect_type, cmp_redirect.redirect_type)
            self.assertEqual(t_redirect.pass_query_string, cmp_redirect.pass_query_string) 
            self.assertEqual(t_redirect.replace_key_with, cmp_redirect.replace_key_with)   
            self.assertEqual(t_redirect.replace_key_prefix_with, cmp_redirect.replace_key_prefix_with)  
            self.assertEqual(t_redirect.proto, cmp_redirect.proto)   
            self.assertEqual(t_redirect.host_name, cmp_redirect.host_name)   
            self.assertEqual(t_redirect.http_redirect_code, cmp_redirect.http_redirect_code)   
            self.bucket.delete_bucket_website()

    def test_unnormal_routing_rule(self):
        index_file = 'index.html'
        error_file = 'error.html'

        # correct test
        # correct condition
        include_header1= ConditionInlcudeHeader('host', 'test.oss-cn-beijing-internal.aliyuncs.com')
        include_header2 = ConditionInlcudeHeader('host', 'test.oss-cn-shenzhen-internal.aliyuncs.com')
        condition = Condition(key_prefix_equals='~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])

        # correct redirect
        redirect = Redirect(redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='http://www.test.com/', mirror_pass_query_string=True, 
                        mirror_follow_redirect=True, mirror_check_md5=False, mirror_headers=None)

        # correct routing rules 
        rule_num = 1
        rule = RoutingRule(rule_num, condition=condition, redirect=redirect)
        website_set = BucketWebsite(index_file, error_file, [rule])

        # correct set get delete function
        self.bucket.put_bucket_website(website_set)
        wait_meta_sync()
        website_get = self.bucket.get_bucket_website()
        self.bucket.delete_bucket_website()

        # start incorrect test
        # rule_num type is not int
        self.assertRaises(oss2.exceptions.ClientError, RoutingRule, rule_num='str', condition=condition, redirect=redirect)
        
        # rule_num < 0
        self.assertRaises(oss2.exceptions.ClientError, RoutingRule, rule_num=-1, condition=condition, redirect=redirect)

        # rule_num is none
        self.assertRaises(oss2.exceptions.ClientError, RoutingRule, rule_num=None, condition=condition, redirect=redirect)

        # condition is none
        self.assertRaises(oss2.exceptions.ClientError, RoutingRule, rule_num=1, condition=None, redirect=redirect)
        
        # redirect is none
        self.assertRaises(oss2.exceptions.ClientError, RoutingRule, rule_num=1, condition=condition, redirect=None)

        # condition.http_err_code_return_equals is not 404 when reidrect type is mirror
        t_condition = Condition(key_prefix_equals='~!@#$%^&*()-_=+|\\[]{}<>,./?`~',
                            http_err_code_return_equals=400, include_header_list=[include_header1, include_header2])
        self.assertRaises(oss2.exceptions.ClientError, RoutingRule, rule_num=1, condition=t_condition, redirect=redirect)

        # rules type is not list
        self.assertRaises(oss2.exceptions.ClientError, BucketWebsite, index_file, error_file, rule)
        
        # rules capacity > 5
        self.assertRaises(oss2.exceptions.ClientError, BucketWebsite, index_file, error_file, [rule, rule, rule, rule, rule, rule])

    def test_unnormal_rule_conditon(self):

        # correct condition
        include_header1= ConditionInlcudeHeader('host', 'test.oss-cn-beijing-internal.aliyuncs.com')
        include_header2 = ConditionInlcudeHeader('host', 'test.oss-cn-shenzhen-internal.aliyuncs.com')
        condition = Condition(key_prefix_equals='test',http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])

        # start incorrect conditon test
        # condition include_header_list type is not list
        self.assertRaises(oss2.exceptions.ClientError, Condition, 'test',404, 'str')
        
        # condition include_header_list capacity >5
        self.assertRaises(oss2.exceptions.ClientError, Condition, 'test', 404, 
                        [include_header1, include_header1, include_header1, include_header1, include_header1, include_header1])

    def test_unnormal_redirect_mirror(self):
    
        index_file = 'index.html'
        error_file = 'error.html'

        # correct test
        # correct condition
        include_header1= ConditionInlcudeHeader('host', 'test.oss-cn-beijing-internal.aliyuncs.com')
        include_header2 = ConditionInlcudeHeader('host', 'test.oss-cn-shenzhen-internal.aliyuncs.com')
        condition1 = Condition(key_prefix_equals='1~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])
        condition2 = Condition(key_prefix_equals='2~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])

        # correct redirect
        # mirro_url_slave is indicated
        redirect1 = Redirect(redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='http://www.test.com/',
                            mirror_url_slave='https://www.slave.com/', mirror_url_probe='http://www.test.com/index.html', 
                            mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False) 

        # mirro_url_slave is not indicated
        redirect2 = Redirect(redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='http://www.test.com/',
                            mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # correct routing rule
        rule = RoutingRule(rule_num=1, condition=condition1, redirect=redirect1)
        rule = RoutingRule(rule_num=2, condition=condition2, redirect=redirect2)
        website_set = BucketWebsite(index_file, error_file, [rule])

        # correct set & get & delete function 
        self.bucket.put_bucket_website(website_set)
        wait_meta_sync()
        website_get = self.bucket.get_bucket_website()
        self.bucket.delete_bucket_website()

        # start incorrect test        
        # invalid redirect type
        self.assertRaises(oss2.exceptions.ClientError, Redirect,
                        redirect_type='unnormal-direct-type', mirror_url='http://www.test.com/', mirror_pass_query_string=True, 
                        mirror_follow_redirect=True, mirror_check_md5=False)

        # mirror_url not startwith 'http' or 'htttps'
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='www.test.com/', 
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # mirror_url not endwith '/'                  
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='http://www.test.com', 
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # mirror_url_slave not startwith 'http' or 'htttps'
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='http://www.test.com/', 
                        mirror_url_slave='www.slave.com/', mirror_url_probe='http://www.test.com/index.html',
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # mirror_url_slave not endwith '/'             
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='http://www.test.com/', 
                        mirror_url_slave='https://www.slave.com', mirror_url_probe='http://www.test.com/index.html',
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # mirror_url_probe is none when mirror_url_slave is diticated            
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, mirror_url='http://www.test.com/', 
                        mirror_url_slave='https://www.slave.com/', mirror_url_probe=None,
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # proto is not empty when reidrect type is mirrorr
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, proto='http', mirror_url='http://www.test.com/', 
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # host_name is not empty when reidrect type is mirrorr
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, host_name='oss.aliyuncs.com', mirror_url='http://www.test.com/', 
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        #replace_key_prefix_with is not empty when reidrect type is mirrorr
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, replace_key_prefix_with='abc',mirror_url='http://www.test.com/',
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)

        # http_redirect_code is not empty when reidrect type is mirrorr
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_MIRROR, http_redirect_code=302, mirror_url='http://www.test.com/', 
                        mirror_pass_query_string=True, mirror_follow_redirect=True, mirror_check_md5=False)   

        # correct pass_list
        mirror_header = RedirectMirrorHeaders(pass_list=['a','b'])
        # pass_list type is not list
        self.assertRaises(oss2.exceptions.ClientError, RedirectMirrorHeaders, pass_list='a')
        # pass_list capacity > 10
        pass_list = ['key1', 'key2', 'key3', 'key4', 'key5', 'key6', 'key7', 'key8', 'key9', 'key10', 'key11']
        self.assertRaises(oss2.exceptions.ClientError, RedirectMirrorHeaders, pass_list=pass_list)

        # correct remove_list
        mirror_header = RedirectMirrorHeaders(remove_list=['a','b'])
        # remove_list type is not list
        self.assertRaises(oss2.exceptions.ClientError, RedirectMirrorHeaders, remove_list='str')
        # remove_list capacity > 10
        remove_list = ['key1', 'key2', 'key3', 'key4', 'key5', 'key6', 'key7', 'key8', 'key9', 'key10', 'key11']
        self.assertRaises(oss2.exceptions.ClientError, RedirectMirrorHeaders, remove_list=remove_list)

        # correct set_list
        t_set = MirrorHeadersSet('key1', 'value1')
        mirror_header = RedirectMirrorHeaders(set_list=[t_set])
        # mirror_header set_list type is not list
        self.assertRaises(oss2.exceptions.ClientError, RedirectMirrorHeaders, set_list=t_set)
        # mirror_header set_list capacity > 10
        set_list = [t_set, t_set, t_set, t_set, t_set, t_set, t_set, t_set, t_set, t_set, t_set]
        self.assertRaises(oss2.exceptions.ClientError, RedirectMirrorHeaders, set_list=set_list)

    def test_unnormal_redirect_alicdn_external_internal(self):
        index_file = 'index.html'
        error_file = 'error.html'

        # correct test
        # correct condition
        include_header1= ConditionInlcudeHeader('host', 'test.oss-cn-beijing-internal.aliyuncs.com')
        include_header2 = ConditionInlcudeHeader('host', 'test.oss-cn-shenzhen-internal.aliyuncs.com')
        condition1 = Condition(key_prefix_equals='1~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])
        condition2 = Condition(key_prefix_equals='2~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])
        condition3 = Condition(key_prefix_equals='3~!@#$%^&*()-_=+|\\[]{}<>,./?`~', 
                            http_err_code_return_equals=404, include_header_list=[include_header1, include_header2])

        # correct reidrect AliCDN 
        redirect1 = Redirect(redirect_type=REDIRECT_TYPE_ALICDN, pass_query_string=True, proto='http', 
                            replace_key_with='${key}.suffix', host_name='oss.aliyuncs.com', http_redirect_code=302)

        # correct reidrect  External
        redirect2 = Redirect(redirect_type=REDIRECT_TYPE_EXTERNAL, pass_query_string=False, proto='https', 
                            replace_key_prefix_with='abc', host_name='oss.aliyuncs.com', http_redirect_code=302)

        # correct reidrect  Internal
        redirect3 = Redirect(redirect_type=REDIRECT_TYPE_INTERNAL, pass_query_string=False, replace_key_with='${key}.suffix')

        # correct rules
        rule1 = RoutingRule(rule_num=1, condition=condition1, redirect=redirect1)
        rule2 = RoutingRule(rule_num=2, condition=condition2, redirect=redirect2)
        rule3 = RoutingRule(rule_num=3, condition=condition3, redirect=redirect3)

        # correct website set get and delete function
        website_set = BucketWebsite(index_file, error_file, [rule1, rule2, rule3])
        self.bucket.put_bucket_website(website_set)
        wait_meta_sync()
        website_get = self.bucket.get_bucket_website()
        self.bucket.delete_bucket_website()

        # start incorrect test
        # http_redirect_code < 300
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_ALICDN, pass_query_string=True, replace_key_with='${key}.suffix', proto='http', 
                        host_name='oss.aliyuncs.com', http_redirect_code=299)

        # http_redirect_code > 399
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_ALICDN, pass_query_string=True, replace_key_with='${key}.suffix', proto='http',  
                        host_name='oss.aliyuncs.com', http_redirect_code=400)     

        # replace_key_with and replace_key_prefix_with both is exsit
        self.assertRaises(oss2.exceptions.ClientError, Redirect, 
                        redirect_type=REDIRECT_TYPE_ALICDN, pass_query_string=True, replace_key_with='${key}.suffix', proto='http', 
                        host_name='oss.aliyuncs.com', replace_key_prefix_with='abc', http_redirect_code=302)

        # host_name is not empty when reidirect type is Internal
        self.assertRaises(oss2.exceptions.ClientError,Redirect,
                        redirect_type=REDIRECT_TYPE_INTERNAL, pass_query_string=False, replace_key_with='${key}.suffix', host_name='oss.aliyuncs.com')

        # proto is not empty when redirect type is Internal
        self.assertRaises(oss2.exceptions.ClientError,Redirect,
                        redirect_type=REDIRECT_TYPE_INTERNAL, pass_query_string=False, replace_key_with='${key}.suffix', proto='http')

        # http_redirect_code is not empty when reidirect type is Internal
        self.assertRaises(oss2.exceptions.ClientError,Redirect,
                        redirect_type=REDIRECT_TYPE_INTERNAL, pass_query_string=False, replace_key_with='${key}.suffix', http_redirect_code=302)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import requests
import filecmp
import calendar
import csv
import re
import sys

from oss2.exceptions import (ClientError, RequestError, NoSuchBucket,
                             NotFound, NoSuchKey, Conflict, PositionNotEqualToLength, ObjectNotAppendable, SelectOperationFailed, SelectOperationClientError)
from .common import *

from oss2.select_response import SelectResponseAdapter

if sys.version_info[0] > 2:
    # py3k
    def _open(file):
        return open(file, 'r', encoding='utf-8')
else:
    # py2
    def _open(file):
        return open(file, 'r')

def now():
    return int(calendar.timegm(time.gmtime()))

class SelectCsvObjectTestHelper(object):
    def __init__(self, bucket):
        self.bucket = bucket
        self.scannedSize = 0

    def select_call_back(self, consumed_bytes, total_bytes = None):
        self.scannedSize = consumed_bytes

    def test_select_csv_object(self, testCase, sql, line_range = None):
        key = "city_sample_data.csv"
        result = self.bucket.put_object_from_file(key, 'tests/sample_data.csv')
        result = self.bucket.create_select_object_meta(key)
        result = self.bucket.head_object(key)
        file_size = result.content_length
        input_format = {'CsvHeaderInfo' : 'Use'}
        if (line_range is not None):
            input_format['LineRange'] = line_range
        
        SelectResponseAdapter._FRAMES_FOR_PROGRESS_UPDATE = 0
        result = self.bucket.select_object(key, sql, self.select_call_back, input_format)
        content = b''
        for chunk in result:
            content += chunk
        
        print(result.request_id)
        testCase.assertEqual(result.status, 206, result.request_id)
        testCase.assertTrue(len(content) > 0)

        if line_range is None:
            testCase.assertEqual(self.scannedSize, file_size)

        return content

    def test_select_csv_object_invalid_request(self, testCase, sql, line_range = None):
        key = "city_sample_data.csv"
        self.bucket.put_object_from_file(key, 'tests/sample_data.csv')

        result = self.bucket.create_select_object_meta(key)
        file_size = result.content_length

        input_format = {'CsvHeaderInfo' : 'Use'}
        if (line_range is not None):
            input_format['Range'] = line_range

        try:
            result = self.bucket.select_object(key, sql, None, input_format)
            testCase.assertEqual(result.status, 400)
        except oss2.exceptions.ServerError as e:
            testCase.assertEqual(e.status, 400)

class TestSelectCsvObject(OssTestCase):
    def test_select_csv_object_not_empty_city(self):
        helper = SelectCsvObjectTestHelper(self.bucket)
        content = helper.test_select_csv_object(self, "select Year, StateAbbr, CityName, PopulationCount from ossobject where CityName != ''")
       
        with _open('tests/sample_data.csv') as csvfile:
            spamreader = csv.DictReader(csvfile, delimiter=',', quotechar='"')
            select_data = b''
            for row in spamreader:
                line = b''
                if row['CityName'] != '':
                        line += row['Year'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['StateAbbr'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['CityName'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['PopulationCount'].encode('utf-8')
                        line += '\n'.encode('utf-8')
                        select_data += line
            
            self.assertEqual(select_data, content)
    
    def test_select_csv_object_like(self):
        helper = SelectCsvObjectTestHelper(self.bucket) 
        content = helper.test_select_csv_object(self, "select Year, StateAbbr, CityName, Short_Question_Text from ossobject where Measure like '%blood pressure%Years'")

        with _open('tests/sample_data.csv') as csvfile:
            spamreader = csv.DictReader(csvfile, delimiter=',', quotechar='"')
            select_data = b''
            matcher = re.compile('^.*blood pressure.*Years$')

            for row in spamreader:
                line = b''
                if matcher.match(row['Measure']):
                        line += row['Year'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['StateAbbr'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['CityName'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['Short_Question_Text'].encode('utf-8')
                        line += '\n'.encode('utf-8')
                        select_data += line
            
            self.assertEqual(select_data, content)
        
    def test_select_csv_object_line_range(self):
        helper = SelectCsvObjectTestHelper(self.bucket) 
        content = helper.test_select_csv_object(self, "select Year,StateAbbr, CityName, Short_Question_Text from ossobject'", (0, 50))

        with _open('tests/sample_data.csv') as csvfile:
            spamreader = csv.DictReader(csvfile, delimiter=',', quotechar='"')
            select_data = b''
            count = 0
            for row in spamreader:
                if count < 50:
                    line = b''
                    line += row['Year'].encode('utf-8')
                    line += ','.encode('utf-8')
                    line += row['StateAbbr'].encode('utf-8')
                    line += ','.encode('utf-8')
                    line += row['CityName'].encode('utf-8')
                    line += ','.encode('utf-8')
                    line += row['Short_Question_Text'].encode('utf-8')
                    line += '\n'.encode('utf-8')
                    select_data += line 
                else:
                    break
                count += 1
            
            self.assertEqual(select_data, content)
    
    def test_select_csv_object_int_aggregation(self):
        helper = SelectCsvObjectTestHelper(self.bucket) 
        content = helper.test_select_csv_object(self, "select avg(cast(year as int)), max(cast(year as int)), min(cast(year as int)) from ossobject where year = 2015")
        self.assertEqual(content, b'2015,2015,2015\n')
    
    def test_select_csv_object_float_aggregation(self):
        helper = SelectCsvObjectTestHelper(self.bucket) 
        content = helper.test_select_csv_object(self, "select avg(cast(data_value as double)), max(cast(data_value as double)), sum(cast(data_value as double)) from ossobject")
        # select_data = b''

        with _open('tests/sample_data.csv') as csvfile:
            spamreader = csv.DictReader(csvfile, delimiter=',', quotechar='"')
            sum = 0.0
            avg = 0.0
            line_count = 0
            max = 0.0
            for row in spamreader:
                if len(row['Data_Value']) > 0 :
                    val = float(row['Data_Value'])
                    if val > max:
                        max = val
                
                    sum += val
                    line_count += 1
            
            avg = sum/line_count
            # select_data = ("{0:.4f}".format(avg) + "," + str(max) + "," + "{0:.1f}".format(sum) + '\n').encode('utf-8')
            aggre_results = content.split(b',')
            avg_result = float(aggre_results[0])
            max_result = float(aggre_results[1])
            sum_result = float(aggre_results[2])
            self.assertEqual(avg, avg_result)
            self.assertEqual(max, max_result)
            self.assertEqual(sum, sum_result)

    def test_select_csv_object_concat(self):
        helper = SelectCsvObjectTestHelper(self.bucket) 
        content = helper.test_select_csv_object(self, "select Year,StateAbbr, CityName, Short_Question_Text from ossobject where (data_value || data_value_unit) = '14.8%'")
        select_data = b''

        with _open('tests/sample_data.csv') as csvfile:
            spamreader = csv.DictReader(csvfile, delimiter=',', quotechar='"')
            for row in spamreader:
                line = b''
                if row['Data_Value_Unit'] == '%' and row['Data_Value'] == '14.8' :
                        line += row['Year'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['StateAbbr'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['CityName'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['Short_Question_Text'].encode('utf-8')
                        line += '\n'.encode('utf-8')
                        select_data += line
        
            self.assertEqual(select_data, content)

    def test_select_csv_object_complicate_condition(self):
        helper = SelectCsvObjectTestHelper(self.bucket) 
        content = helper.test_select_csv_object(self, "select Year,StateAbbr, CityName, Short_Question_Text, data_value, data_value_unit, category, high_confidence_limit from ossobject where data_value > 14.8 and data_value_unit = '%' or Measure like '%18 Years' and Category = 'Unhealthy Behaviors' or high_confidence_limit > 70.0 ")
        select_data = b''

        matcher = re.compile('^.*18 Years$')
        with _open('tests/sample_data.csv') as csvfile:
            spamreader = csv.DictReader(csvfile, delimiter=',', quotechar='"')
            for row in spamreader:
                line = b''
                if len(row['Data_Value']) > 0 and float(row['Data_Value']) > 14.8 and row['Data_Value_Unit'] == '%' or matcher.match(row['Measure']) and row['Category'] == 'Unhealthy Behaviors' or len(row['High_Confidence_Limit']) > 0 and float(row['High_Confidence_Limit']) > 70.0 :
                        line += row['Year'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['StateAbbr'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['CityName'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['Short_Question_Text'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['Data_Value'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['Data_Value_Unit'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['Category'].encode('utf-8')
                        line += ','.encode('utf-8')
                        line += row['High_Confidence_Limit'].encode('utf-8')
                        line += '\n'.encode('utf-8')
                        select_data += line
            self.assertEqual(select_data, content)
    
    def test_select_csv_object_invalid_sql(self):
        helper = SelectCsvObjectTestHelper(self.bucket) 

        helper.test_select_csv_object_invalid_request(self, "select * from ossobject where avg(cast(year as int)) > 2016")
        helper.test_select_csv_object_invalid_request(self, "")
        helper.test_select_csv_object_invalid_request(self, "select year || CityName from ossobject")
        helper.test_select_csv_object_invalid_request(self, "select * from ossobject group by CityName")
        helper.test_select_csv_object_invalid_request(self, "select * from ossobject order by _1")
        helper.test_select_csv_object_invalid_request(self, "select * from ossobject oss join s3object s3 on oss.CityName = s3.CityName")

    def test_select_csv_object_with_invalid_data(self):
        key = "invalid_city_sample_data.csv"
        self.bucket.put_object_from_file(key, 'tests/invalid_sample_data.csv')
        input_format = {'CsvHeaderInfo' : 'Use'}
        result = self.bucket.select_object(key, "select _1 from ossobject", None, input_format)
        content = b''

        try:
            for chunk in result:
                content += chunk
            self.assertFalse(True, "expect to raise exception")
        except SelectOperationFailed:
            print("Got the exception. Ok.")

    def test_select_csv_object_into_file(self):
        key = "city_sample_data.csv"
        self.bucket.put_object_from_file(key, 'tests/sample_data.csv')
        input_format = {'CsvHeaderInfo' : 'None',
                        'CommentCharacter' : '#',
                        'RecordDelimiter' : '\n',
                        'FieldDelimiter' : ',',
                        'QuoteCharacter' : '"',
                        'SplitRange' : (0, None)
                        }
        output_file = 'tests/sample_data_out.csv'

        head_csv_params = {  
                            'RecordDelimiter' : '\n',
                            'FieldDelimiter' : ',',
                            'QuoteCharacter' : '"',
                            'OverwriteIfExists': True}

        self.bucket.create_select_object_meta(key, head_csv_params)

        self.bucket.select_object_to_file(key, output_file, "select * from ossobject", None, input_format)
        f1 = _open('tests/sample_data.csv')
        content1 = f1.read()
        f1.close()

        f2 = _open(output_file)
        content2 = f2.read()
        f2.close()

        os.remove(output_file)
        self.assertEqual(content1, content2)

    def test_select_gzip_csv_object_into_file(self):
        key = "city_sample_data.csv.gz"
        self.bucket.put_object_from_file(key, 'tests/sample_data.csv.gz')
        input_format = {'CsvHeaderInfo' : 'None',
                        'CommentCharacter' : '#',
                        'RecordDelimiter' : '\n',
                        'FieldDelimiter' : ',',
                        'QuoteCharacter' : '"',
                        'CompressionType' : 'GZIP'
                        }
        output_file = 'tests/sample_data_out.csv'

        self.bucket.select_object_to_file(key, output_file, "select * from ossobject", None, input_format)
        f1 = _open('tests/sample_data.csv')
        content1 = f1.read()
        f1.close()

        f2 = _open(output_file)
        content2 = f2.read()
        f2.close()

        os.remove(output_file)
        self.assertEqual(content1, content2)

    def test_select_csv_object_none_range(self):
        key = "city_sample_data.csv"
        self.bucket.put_object_from_file(key, 'tests/sample_data.csv')
        self.bucket.create_select_object_meta(key)

        input_formats = [
                            { 'SplitRange' : (None, None) },
                            { 'LineRange' : (None, None) },
                            {'SplitRange' : None },
                            {'LineRange' : None }
                        ]

        for input_format in input_formats:
            result = self.bucket.select_object(key, "select * from ossobject", None, input_format)
            content = b''
            for chunk in result:
                content += chunk

            self.assertTrue(len(content) > 0)

    def test_with_select_result(self):
        key = "city_sample_data.csv"
        self.bucket.put_object_from_file(key, 'tests/sample_data.csv')
        self.bucket.create_select_object_meta(key)

        input_formats = [
            {'SplitRange': (None, None)},
            {'LineRange': (None, None)},
            {'SplitRange': None},
            {'LineRange': None}
        ]

        for input_format in input_formats:
            result1 = self.bucket.select_object(key, "select * from ossobject", None, input_format)
            content1 = b''
            for chunk in result1:
                content1 += chunk
            self.assertTrue(len(content1) > 0)

            result2 = self.bucket.select_object(key, "select * from ossobject", None, input_format)
            content2 = b''
            with result2 as f:
                for chunk in f:
                    content2 += chunk
            self.assertEqual(content1, content2)

            result3 = self.bucket.select_object(key, "select * from ossobject", None, input_format)
            with result3 as f:
                pass
            content3 = result3.read()
            self.assertEqual(0, len(content3))

    def test_create_csv_object_meta_invalid_request(self):
        key = "city_sample_data.csv"
        self.bucket.put_object_from_file(key, 'tests/sample_data.csv')
        format = {'CompressionType':'GZIP'}
        try:
            self.bucket.create_select_object_meta(key, format)
            self.assertFalse(True, "expected error did not occur")
        except oss2.exceptions.ServerError :
            print("expected error occured")

    def test_create_csv_object_meta_invalid_request2(self):
        key = "city_sample_data.csv"
        self.bucket.put_object_from_file(key, 'tests/sample_data.csv')
        format = {'invalid_type':'value', 'CompressionType':'GZIP'}
        try:
            self.bucket.create_select_object_meta(key, format)
            self.assertFalse(True, "expected error did not occur")
        except SelectOperationClientError:
            print("expected error occured")

    def test_select_csv_object_with_output_delimiters(self):
        key = "test_select_csv_object_with_output_delimiters"
        content = "abc,def\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'OutputRecordDelimiter':'\r\n', 'OutputFieldDelimiter':'|'}
        result = self.bucket.select_object(key, "select _1, _2 from ossobject", None, select_params)
        content = b''
        for chunk in result:
            content += chunk
        
        self.assertEqual(content, 'abc|def\r\n'.encode('utf-8'))

    def test_select_csv_object_with_crc(self):
        key = "test_select_csv_object_with_crc"
        content = "abc,def\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'EnablePayloadCrc':True}
        result = self.bucket.select_object(key, "select * from ossobject where true", None, select_params)
        content = result.read()
        
        self.assertEqual(content, 'abc,def\n'.encode('utf-8'))
    
    def test_select_csv_object_with_skip_partial_data(self):
        key = "test_select_csv_object_with_skip_partial_data"
        content = "abc,def\nefg\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'SkipPartialDataRecord':'true'}
        result = self.bucket.select_object(key, "select _1, _2 from ossobject", None, select_params)
        content = b''
        try:
            for chunk in result:
                content += chunk
            self.assertFalse("expected error did not occur")
        except SelectOperationFailed:
            print("expected error occurs")
        
        self.assertEqual(content, 'abc,def\n'.encode('utf-8'))

    def test_select_csv_object_with_max_partial_data(self):
        key = "test_select_csv_object_with_skip_partial_data"
        content = "abc,def\nefg\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'SkipPartialDataRecord':'true', "MaxSkippedRecordsAllowed":100}
        result = self.bucket.select_object(key, "select _1, _2 from ossobject", None, select_params)
        content = b''
        for chunk in result:
            content += chunk
        
        self.assertEqual(content, 'abc,def\n'.encode('utf-8'))

    def test_select_csv_object_with_output_raw(self):
        key = "test_select_csv_object_with_output_raw"
        content = "abc,def\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'OutputRawData':'true'}
        result = self.bucket.select_object(key, "select _1 from ossobject", None, select_params)
        content = b''
        for chunk in result:
            content += chunk
        
        self.assertEqual(content, 'abc\n'.encode('utf-8'))
    
    def test_select_csv_object_with_keep_columns(self):
        key = "test_select_csv_object_with_keep_columns"
        content = "abc,def\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'KeepAllColumns':'true'}
        result = self.bucket.select_object(key, "select _1 from ossobject", None, select_params)
        content = b''
        for chunk in result:
            content += chunk
        
        self.assertEqual(content, 'abc,\n'.encode('utf-8'))
    
    def test_select_csv_object_with_output_header(self):
        key = "test_select_csv_object_with_output_header"
        content = "name,job\nabc,def\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'OutputHeader':'true', 'CsvHeaderInfo':'Use'}
        result = self.bucket.select_object(key, "select name from ossobject", None, select_params)
        content = b''
        for chunk in result:
            content += chunk
        
        self.assertEqual(content, 'name\nabc\n'.encode('utf-8'))
    
    def test_select_csv_object_with_invalid_parameters(self):
        key = "test_select_csv_object_with_invalid_parameters"
        content = "name,job\nabc,def\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'OutputHeader123':'true', 'CsvHeaderInfo':'Use'}
        try:
            result = self.bucket.select_object(key, "select name from ossobject", None, select_params)
            self.assertFalse()
        except SelectOperationClientError:
            print("expected error")
    
    def test_select_csv_object_with_bytes_range(self):
        key = "test_select_csv_object_with_bytes_range"
        content = "test\nabc\n"
        self.bucket.put_object(key, content.encode('utf-8'))
        select_params = {'AllowQuotedRecordDelimiter':False}
        byte_range = [0,1]
        result = self.bucket.select_object(key, "select * from ossobject", None, select_params,  byte_range)
        content = b''
        for chunk in result:
            content += chunk
        self.assertEqual('test\n'.encode('utf-8'), content)

    def test_select_csv_object_with_bytes_range_invalid(self):
        key = "test_select_csv_object_with_bytes_range"
        content = "test\nabc\n"
        self.bucket.put_object(key, content.encode('utf-8'))
        byte_range = [0,1]
        try:
            self.bucket.select_object(key, "select * from ossobject", None, None,  byte_range)
            self.assertFalse()
        except ClientError:
            print("expected error")

        select_params = {'AllowQuotedRecordDelimiter':True}

        try:
            self.bucket.select_object(key, "select * from ossobject", None, select_params,  byte_range)
            self.assertFalse()
        except ClientError:
            print("expected error")
        
        select_params = {'AllowQuotedRecordDelimiter':False, 'Json_Type':'LINES'}

        try:
            self.bucket.select_object(key, "select * from ossobject", None, select_params,  byte_range)
            self.assertFalse()
        except SelectOperationClientError:
            print("expected error")

    
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import requests
import filecmp
import calendar
import json
import re
import sys

from oss2.exceptions import (ClientError, RequestError, NoSuchBucket,
                             NotFound, NoSuchKey, Conflict, PositionNotEqualToLength, ObjectNotAppendable, SelectOperationFailed,SelectOperationClientError)
from .common import *

if sys.version_info[0] > 2:
    # py3k
    def _open(file):
        return open(file, 'r', encoding='utf-8')
else:
    # py2
    def _open(file):
        return open(file, 'r')

def now():
    return int(calendar.timegm(time.gmtime()))

def  select_call_back(consumed_bytes, total_bytes =  None):
	print('Consumed Bytes:'  +  str(consumed_bytes) +  '\n')

class SelectJsonObjectTestHelper(object):
    def __init__(self, bucket):
        self.bucket = bucket
        self.scannedSize = 0

    def select_call_back(self, consumed_bytes, total_bytes = None):
        self.scannedSize = consumed_bytes

    def test_select_json_object(self, testCase, sql, input_format):
        is_gzip = False
        if input_format['Json_Type'] == 'DOCUMENT':
            if 'CompressionType' in input_format and input_format['CompressionType'] == 'GZIP':
                key = "sample_json.json.gz"
                local_file = 'tests/sample_json.json.gz'
                is_gzip = True
            else:
                key = "sample_json.json"
                local_file = 'tests/sample_json.json'
        else:
            key = 'sample_json_lines.json'
            local_file = 'tests/sample_json_lines.json'

        result = self.bucket.put_object_from_file(key, local_file)
        
        #set OutputrecordDelimiter with ',' to make json load the result correctly
        input_format['OutputRecordDelimiter'] = ','

        if (input_format['Json_Type'] == 'LINES'):
            result = self.bucket.create_select_object_meta(key, {'Json_Type':'LINES'})

        result = self.bucket.head_object(key)
        file_size = result.content_length

        result = self.bucket.select_object(key, sql, self.select_call_back, input_format)
        content = b''
        for chunk in result:
            content += chunk
        
        testCase.assertEqual(result.status, 206, result.request_id)
        testCase.assertTrue(len(content) > 0)

        if 'SplitRange' not in input_format and 'LineRange' not in input_format and is_gzip is False:
            testCase.assertEqual(self.scannedSize, file_size)

        return content

    def test_select_json_object_invalid_request(self, testCase, sql, input_format):
        if input_format['Json_Type'] == 'DOCUMENT':
            key = "sample_json.json"
            local_file = 'tests/sample_json.json'
        else:
            key = 'sample_json_lines.json'
            local_file = 'tests/sample_json_lines.json'

        result = self.bucket.put_object_from_file(key, local_file)
        if (input_format['Json_Type'] == 'LINES'):
            result = self.bucket.create_select_object_meta(key, input_format)

        file_size = result.content_length

        try:
            result = self.bucket.select_object(key, sql, None, input_format)
            testCase.assertEqual(result.status, 400)
        except oss2.exceptions.ServerError as e:
            testCase.assertEqual(e.status, 400)

class TestSelectJsonObject(OssTestCase):
    
    def test_select_json_document_democrat_senators(self):
        print("test_select_json_document_democrat_senators")
        helper = SelectJsonObjectTestHelper(self.bucket)
        input_format = {'Json_Type':'DOCUMENT', 'CompressionType':'None'}
        content = helper.test_select_json_object(self, "select * from ossobject.objects[*] where party = 'Democrat'", input_format)
        
        content = content[0:len(content)-1] #remove the last ','
        content = b"[" + content + b"]"  #make json parser happy
        result = json.loads(content.decode('utf-8'))

        result_index = 0
        with _open('tests/sample_json.json') as json_file:
            data = json.load(json_file)
            for row in data['objects']:
                if row['party'] == 'Democrat':
                    self.maxDiff = None
                    self.assertEqual(result[result_index], row, str(result_index))
                    result_index += 1
       
    def test_select_json_lines_democrat_senators(self):
        print("test_select_json_lines_democrat_senators")
        helper = SelectJsonObjectTestHelper(self.bucket)
        input_format = {'Json_Type':'LINES'}
        content = helper.test_select_json_object(self, "select * from ossobject where party = 'Democrat'", input_format)
        
        content = content[0:len(content)-1] #remove the last ','
        content = b"[" + content + b"]"  #make json parser happy
        result = json.loads(content.decode('utf-8'))

        result2 = []
        with _open('tests/sample_json.json') as json_file:
            data = json.load(json_file)
            for row in data['objects']:
                if row['party'] == 'Democrat':
                    result2.append(row)
            
            self.assertEqual(result, result2)
    
    def test_select_json_object_like(self):
        print("test_select_json_object_like")
        helper = SelectJsonObjectTestHelper(self.bucket)
        select_params = {'Json_Type':'LINES'}
        content = helper.test_select_json_object(self, "select person.firstname, person.lastname from ossobject where person.birthday like '1959%'", select_params)
        content = content[0:len(content)-1] #remove the last ','
        content = b"[" + content + b"]"  #make json parser happy
        result = json.loads(content.decode('utf-8'))

        index = 0
        with _open('tests/sample_json.json') as json_file:
            data = json.load(json_file)
            for row in data['objects']:
                select_row = {}
                if row['person']['birthday'].startswith('1959'): 
                        select_row['firstname'] = row['person']['firstname']
                        select_row['lastname'] = row['person']['lastname']
                        self.assertEqual(result[index], select_row)
                        index += 1

    def test_select_gzip_json_object_like(self):
        print("test_select_gzip_json_object_like")
        helper = SelectJsonObjectTestHelper(self.bucket)
        select_params = {'Json_Type':'DOCUMENT', 'CompressionType':'GZIP'}
        content = helper.test_select_json_object(self, "select person.firstname, person.lastname from ossobject.objects[*]  where person.birthday like '1959%'", select_params)
        content = content[0:len(content)-1] #remove the last ','
        content = b"[" + content + b"]"  #make json parser happy
        result = json.loads(content.decode('utf-8'))

        index = 0
        with _open('tests/sample_json.json') as json_file:
            data = json.load(json_file)
            for row in data['objects']:
                select_row = {}
                if row['person']['birthday'].startswith('1959'): 
                        select_row['firstname'] = row['person']['firstname']
                        select_row['lastname'] = row['person']['lastname']
                        self.assertEqual(result[index], select_row)
                        index += 1

    def test_select_json_object_with_output_raw(self):
        key = "test_select_json_object_with_output_raw"
        content = "{\"key\":\"abc\"}"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'OutputRawData':'true', 'Json_Type':'DOCUMENT'}
        result = self.bucket.select_object(key, "select key from ossobject", None, select_params)
        content = b''
        for chunk in result:
            content += chunk
        
        self.assertEqual(content, '{\"key\":\"abc\"}\n'.encode('utf-8'))

    def test_select_json_object_with_crc(self):
        key = "test_select_json_object_with_crc"
        content = "{\"key\":\"abc\"}\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'EnablePayloadCrc':True, 'Json_Type':'DOCUMENT'}
        result = self.bucket.select_object(key, "select * from ossobject where true", None, select_params)
        content = result.read()
        
        self.assertEqual(content, '{\"key\":\"abc\"}\n'.encode('utf-8'))

    def test_select_json_object_with_skip_partial_data(self):
        key = "test_select_json_object_with_skip_partial_data"
        content = "{\"key\":\"abc\"},{\"key2\":\"def\"}"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'SkipPartialDataRecord':'false', 'Json_Type':'LINES', 'MaxSkippedRecordsAllowed':100}
        result = self.bucket.select_object(key, "select key from ossobject", None, select_params)
        content = b''
        try:
            for chunk in result:
                content += chunk
        except oss2.exceptions.ServerError:
            print("expected error occurs")
        
        self.assertEqual(content, '{\"key\":\"abc\"}\n{}\n'.encode('utf-8'))
    
    def test_select_json_object_with_invalid_parameter(self):
        key = "test_select_json_object_with_invalid_parameter"
        content = "{\"key\":\"abc\"}\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'EnablePayloadCrc':True, 'Json_Type':'DOCUMENT', 'unsupported_param':True}

        try:
            self.bucket.select_object(key, "select * from ossobject where true", None, select_params)
            self.assertFalse(True, 'expected error did not occur')
        except SelectOperationClientError:
            print("expected error occurs")

    def test_create_json_meta_with_invalid_parameter(self):
        key = "test_create_json_meta_with_invalid_parameter"
        content = "{\"key\":\"abc\"}\n"
        self.bucket.put_object(key, content.encode('utf_8'))
        select_params = {'EnablePayloadCrc':True, 'Json_Type':'DOCUMENT'}

        try:
            self.bucket.create_select_object_meta(key,select_params)
            self.assertFalse(True, 'expected error did not occur')
        except SelectOperationClientError:
            print("expected error occurs")

    def test_create_json_object_meta_invalid_request2(self):
        key = "sample_json.json"
        self.bucket.put_object_from_file(key, 'tests/sample_json.json')
        format = {'Json_Type':'invalid', 'CompressionType':'None', 'OverwriteifExists':'True'}
        try:
            self.bucket.create_select_object_meta(key, format)
            self.assertFalse(True, "expected error did not occur")
        except SelectOperationClientError:
            print("expected error occured")

    def test_select_json_object_line_range(self):
        print("test_select_json_object_line_range")
        helper = SelectJsonObjectTestHelper(self.bucket) 

        select_params = {'LineRange' : (10, 50), 'Json_Type':'LINES'}
        content = helper.test_select_json_object(self, "select person.firstname as aaa as firstname, person.lastname, extra from ossobject'", select_params)
        content = content[0:len(content)-1] #remove the last ','
        content = b"[" + content + b"]"  #make json parser happy
        result = json.loads(content.decode('utf-8'))

        result_index = 0
        result2 = []
        with _open('tests/sample_json.json') as json_file:
            data = json.load(json_file)
            index = 0
            for row in data['objects']:
                select_row = {}
                if index >= 10 and index < 50:
                    select_row['firstname'] = row['person']['firstname']
                    select_row['lastname'] = row['person']['lastname']
                    select_row['extra'] = row['extra']
                    self.assertEqual(result[result_index], select_row)
                    result_index += 1
                elif index >= 50:
                    break
                index += 1
        
    def test_select_json_object_int_aggregation(self):
        print("test_select_json_object_int_aggregation")
        helper = SelectJsonObjectTestHelper(self.bucket) 
        select_params = {'Json_Type': 'DOCUMENT'}
        content = helper.test_select_json_object(self, "select avg(cast(person.cspanid as int)), max(cast(person.cspanid as int)), min(cast(person.cspanid as int)) from ossobject.objects[*] where person.cspanid = 1011723", select_params)
        self.assertEqual(content, b'{\"_1\":1011723,\"_2\":1011723,\"_3\":1011723},')
    
    def test_select_json_object_float_aggregation(self):
        print("test_select_json_object_float_aggregation")
        helper = SelectJsonObjectTestHelper(self.bucket) 
        select_params = {'Json_Type': 'DOCUMENT'}
        content = helper.test_select_json_object(self, "select avg(cast(person.cspanid as double)), max(cast(person.cspanid as double)), min(cast(person.cspanid as double)) from ossobject.objects[*]", select_params)
        print(content)
        
    def test_select_json_object_concat(self):
        print("test_select_json_object_concat")
        helper = SelectJsonObjectTestHelper(self.bucket) 
        select_params = {'Json_Type': 'DOCUMENT'}
        content = helper.test_select_json_object(self, "select person from ossobject.objects[*] where (person.firstname || person.lastname) = 'JohnKennedy'", select_params)
        content = content[0:len(content)-1] #remove the last ','
        content = b"[" + content + b"]"  #make json parser happy
        result = json.loads(content.decode('utf-8'))

        result_index = 0
        with _open('tests/sample_json.json') as json_file:
            data = json.load(json_file)
            for row in data['objects']:
                if row['person']['firstname'] == 'John' and row['person']['lastname'] == 'Kennedy' :
                    self.assertEqual(result[result_index]['person'], row['person'])
                    result_index += 1
       
    def test_select_json_object_complicate_condition(self):
        print("test_select_json_object_complicate_condition")
        helper = SelectJsonObjectTestHelper(self.bucket) 
        select_params = {'Json_Type' : 'LINES'}
        content = helper.test_select_json_object(self, "select person.firstname, person.lastname, congress_numbers from ossobject where startdate > '2017-01-01' and senator_rank = 'junior' or state = 'CA' and party = 'Republican' ", select_params)
        content = content[0:len(content)-1] #remove the last ','
        content = b"[" + content + b"]"  #make json parser happy
        result = json.loads(content.decode('utf-8'))

        result_index = 0
        with _open('tests/sample_json.json') as json_file:
            data = json.load(json_file)
            for row in data['objects']:
                if row['startdate'] > '2017-01-01' and row['senator_rank'] == 'junior' or row['state'] == 'CA' and row['party'] == 'Republican':
                    self.assertEqual(result[result_index]['firstname'], row['person']['firstname'])
                    self.assertEqual(result[result_index]['lastname'], row['person']['lastname'])
                    self.assertEqual(result[result_index]['congress_numbers'], row['congress_numbers'])
                    result_index += 1
    
    def test_select_json_object_invalid_sql(self):
        print("test_select_json_object_invalid_sql")
        helper = SelectJsonObjectTestHelper(self.bucket) 

        select_params = {'Json_Type':'LINES'}
        helper.test_select_json_object_invalid_request(self, "select * from ossobject where avg(cast(person.birthday as int)) > 2016", select_params)
        helper.test_select_json_object_invalid_request(self, "", select_params)
        helper.test_select_json_object_invalid_request(self, "select person.lastname || person.firstname from ossobject", select_params)
        helper.test_select_json_object_invalid_request(self, "select * from ossobject group by person.firstname", select_params)
        helper.test_select_json_object_invalid_request(self, "select * from ossobject order by _1", select_params)
        helper.test_select_json_object_invalid_request(self, "select * from ossobject oss join s3object s3 on oss.CityName = s3.CityName", select_params)
    
    def test_select_json_object_with_invalid_data(self):
        print("test_select_json_object_with_invalid_data")
        key = "invalid_json.json"
        self.bucket.put_object_from_file(key, 'tests/invalid_sample_data.csv')
        input_format = {'Json_Type' : 'DOCUMENT'}

        try:
            result = self.bucket.select_object(key, "select _1 from ossobject.objects[*]", None, input_format)
            self.assertFalse(True, "expect to raise exception")
        except oss2.exceptions.ServerError:
            print("Got the exception. Ok.")
    
    def test_select_json_object_none_range(self):
        print("test_select_json_object_none_range")
        key = "sample_json.json"
        self.bucket.put_object_from_file(key, 'tests/sample_json.json')
        select_params = {'Json_Type':'LINES'}
        self.bucket.create_select_object_meta(key, select_params)

        input_formats = [
                            { 'SplitRange' : (None, None) },
                            { 'LineRange' : (None, None) },
                            {'SplitRange' : None },
                            {'LineRange' : None }
                        ]

        for input_format in input_formats:
            input_format['Json_Type'] = 'LINES'
            result = self.bucket.select_object(key, "select * from ossobject", None, input_format)
            content = b''
            for chunk in result:
                content += chunk

            self.assertTrue(len(content) > 0)

    def test_select_json_object_parse_num_as_string(self):
        key = "test_select_json_object_parse_num_as_string"
        self.bucket.put_object(key, b"{\"a\":123456789.123456789}")
        result = self.bucket.select_object(key, "select a from ossobject where cast(a as decimal) = 123456789.1234567890", None, select_params={'ParseJsonNumberAsString': 'true', 'Json_Type':'DOCUMENT'})
        content = b''
        for chunk in result:
            content += chunk
        self.assertEqual(content, b"{\"a\":123456789.123456789}\n")
    
if __name__ == '__main__':
    unittest.main()
